# mc_sim.py
from __future__ import annotations

import os
import numpy as np
import dask
import csv
import json
import cloudpickle
import pathlib
from dask.distributed import Client
from datetime import datetime, timedelta
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from tqdm import tqdm
from typing import Callable, Any, Iterable, Optional
from scipy.stats import rv_continuous, rv_discrete
from monaco.mc_case import Case
from monaco.mc_var import InVar, OutVar, InVarSpace
from monaco.mc_enums import SimFunctions, SampleMethod
from monaco.helper_functions import (get_list, vprint, vwarn, empty_list,
                                     hash_str_repeatable)
from monaco.case_runners import preprocess_case, run_case, postprocess_case
from monaco.tqdm_dask_distributed import tqdm_dask
from monaco.dvars_sensitivity import calc_sensitivities
from monaco.mc_multi_plot import multi_plot_grid_rect


class Sim:
    """
    The main Monte Carlo Simulation object.

    Parameters
    ----------
    name : str
        The name for the simulation.
    ndraws : int
        The number of random draws to perform.
    fcns : dict[monaco.mc_enums.SimFunctions, Callable]
        fcns is a dict with keys SimFunctions.PREPROCESS, RUN, and POSTPROCESS.
        These point to user-defined functions with certain input and output
        structures, please see the documentation on how to construct these
        functions.
    firstcaseismedian : bool, default: False
        Whether the first case represents the median value.
    samplemethod : monaco.mc_enums.SampleMethod, default: 'sobol_random'
        The random sampling method to use.
    seed : int, default: np.random.get_state(legacy=False)['state']['key'][0]
        The random number to seed the simulation.
    singlethreaded : bool, default: False
        Whether to run single threaded rather than using dask.
    daskkwargs : dict, default: dict()
        Kwargs to pass to the dask Client constructor, see:
        https://distributed.dask.org/en/stable/api.html#client
    verbose : bool, default: True
        Whether to print out warning and status messages.
    debug : bool, default: False
        If False, cases that fail while running will be skipped over. If True,
        cases that fail will raise an exception.
    keepsiminput : bool, default: True
        Whether to keep the siminput for each case after running.
    keepsimrawoutput : bool, default: True
        Whether to keep the simrawoutput for each case after postprocessing.
    savesimdata : bool, default: True
        Whether to save the simulation data to disk as a .mcsim file.
    savecasedata : bool, default: True
        Whether to save the full output data for each case to disk as .mccase
        files.
    resultsdir : str | pathlib.Path
        The directory to save simulation and case data to. If None, then this
        defaults to a directory named {name}_results.

    Attributes
    ----------
    rootdir : pathlib.Path
        The directory the simulation was run in.
    filepath : pathlib.Path
        The filepath to the simulation .mcsim datafile.
    invarseeds : list[int]
        The random seeds for each of the input variables.
    outvarseeds : list[int]
        The random seeds for each of the input variables.
    caseseeds : list[int]
        The random seeds for each of the cases.
    inittime : datetime.datetime
        The timestamp when this simulation object was created.
    starttime : datetime.datetime
        The timestamp when the simulation began running.
    endtime : datetime.datetime
        The timestamp when the simulation stopped running.
    runtime : datetime.timedelta
        The length of time it took the simulation to run.
    casespreprocessed : set[int]
        The case numbers which were sucessfully preprocessed.
    casesrun : set[int]
        The case numbers which were sucessfully run.
    casespostprocessed : set[int]
        The case numbers which were sucessfully postprocessed.
    vars : dict[str, monaco.mc_var.InVar]
        All Variables.
    invars : dict[str, monaco.mc_var.InVar]
        The Monte Carlo Input Variables.
    outvars : dict[str, monaco.mc_var.OutVar]
        The Monte Carlo Output Variables.
    constvals : dict[str, Any]
        The constant values to pass to each of the cases.
    cases : list[monaco.mc_case.Case]
        The Monte Carlo Cases.
    ninvars : int
        The number of input variables.
    noutvars : int
        The number of output variables.
    corrcoeffs : numpy.ndarray
        The correlation coefficients between all of the scalar variables.
    covs : numpy.ndarray
        The covariance matrix between all of the scalar variables.
    covvarlist : list[str]
        The names of all the scalar variables.
    runsimid : int
        The unique ID for a particular run of this simulation.
    ncases : int
        The number of cases.
    """
    def __init__(self,
                 name              : str,
                 ndraws            : int,
                 fcns              : dict[SimFunctions, Callable],
                 firstcaseismedian : bool = False,
                 samplemethod      : SampleMethod = SampleMethod.SOBOL_RANDOM,
                 seed              : int  = np.random.get_state(legacy=False)['state']['key'][0],
                 singlethreaded    : bool = True,
                 daskkwargs        : dict = dict(),
                 verbose           : bool = True,
                 debug             : bool = False,
                 keepsiminput      : bool = True,
                 keepsimrawoutput  : bool = True,
                 savesimdata       : bool = False,
                 savecasedata      : bool = False,
                 resultsdir        : str | pathlib.Path | None = None,
                 ) -> None:

        self.checkFcnsInput(fcns)

        self.name = name
        self.verbose = verbose
        self.debug = debug
        self.ndraws = ndraws
        self.fcns = fcns
        self.firstcaseismedian = firstcaseismedian
        self.samplemethod = samplemethod
        self.seed = seed
        self.singlethreaded = singlethreaded
        self.daskkwargs = daskkwargs
        self.keepsiminput = keepsiminput
        self.keepsimrawoutput = keepsimrawoutput
        self.savesimdata = savesimdata
        self.savecasedata = savecasedata

        self.rootdir = pathlib.Path.cwd()
        if isinstance(resultsdir, str):
            resultsdir = pathlib.Path(resultsdir)
        if isinstance(resultsdir, pathlib.Path):
            self.resultsdir = resultsdir
        else:
            self.resultsdir = self.rootdir / f'{self.name}_results'
        if self.savesimdata:
            self.filepath = self.resultsdir / f'{self.name}.mcsim'

        self.invarseeds  : list[int] = []
        self.outvarseeds : list[int] = []
        self.caseseeds   : list[int] = []

        self.inittime  : datetime = datetime.now()
        self.starttime : datetime | None = None
        self.endtime   : datetime | None = None
        self.runtime   : timedelta | None = None

        self.casespreprocessed  : set[int] = set()
        self.casesrun           : set[int] = set()
        self.casespostprocessed : set[int] = set()

        self.vars    : dict[str, InVar | OutVar] = dict()
        self.invars  : dict[str, InVar] = dict()
        self.outvars : dict[str, OutVar] = dict()
        self.constvals : dict[str, Any] = dict()
        self.cases : list[Case] = []
        self.ninvars  : int = 0
        self.noutvars : int = 0

        self.corrcoeffs : np.ndarray | None = None
        self.covs       : np.ndarray | None = None
        self.covvarlist : list[str] | None = None

        self.runsimid : int | None = None

        self.ncases : int = ndraws + 1
        self.setFirstCaseMedian(firstcaseismedian)
        self.setNDraws(self.ndraws)  # will regen runsimid

        self.client = None
        self.cluster = None
        if not self.singlethreaded:
            self.initDaskClient()


    def __del__(self) -> None:
        if self.client is not None:
            self.client.close()


    def __getstate__(self) -> dict:
        """Function for pickling self to save to file."""
        state = self.__dict__.copy()
        state['client'] = None  # don't save cluster to file
        state['cluster'] = None  # don't save cluster to file
        state['cases'] = []  # don't save case data when pickling self
        return state


    def __setstate__(self,
                     state: dict,
                     ) -> None:
        """Function to unpickle self when loading from file."""
        self.__dict__.update(state)
        if self.savecasedata:
            self.loadCases()
        if not self.singlethreaded:
            self.initDaskClient()


    def checkFcnsInput(self,
                       fcns: dict[SimFunctions, Callable],
                       ) -> None:
        """
        Check the `fcns` input dictionary for correctness.

        Parameters
        ----------
        fcns : dict[monaco.mc_enums.SimFunctions, Callable]
            fcns must be a dict with keys SimFunctions.PREPROCESS, RUN, and
            POSTPROCESS, which point to special user-defined functions.
        """
        if set(fcns.keys()) \
           != {SimFunctions.PREPROCESS, SimFunctions.RUN, SimFunctions.POSTPROCESS}:
            raise ValueError(f'Sim argument {fcns=} must have keys ' +
                             f'{SimFunctions.PREPROCESS}, {SimFunctions.RUN}, ' +
                             f'and {SimFunctions.POSTPROCESS}')
        if any(not callable(f) for f in fcns.values()):
            raise ValueError(f'Sim argument {fcns=} must contain functions as values')


    def setFirstCaseMedian(self,
                           firstcaseismedian : bool,
                           ) -> None:
        """
        Make the first case represent the median expected case or not.

        Parameters
        ----------
        firstcaseismedian : bool
            Whether to make the first case the median case.
        """
        if firstcaseismedian:
            self.firstcaseismedian = True
            self.ncases = self.ndraws + 1
        else:
            self.firstcaseismedian = False
            self.ncases = self.ndraws
        if self.invars != dict():
            for invar in self.invars.values():
                invar.setFirstCaseMedian(firstcaseismedian)


    def initDaskClient(self):
        """
        Initialize the dask distributed client.
        """
        if not self.singlethreaded:
            self.client = Client(**self.daskkwargs)
            self.cluster = self.client.cluster

            nworkers = len(self.cluster.workers)
            nthreads = nworkers * self.cluster.worker_spec[0]['options']['nthreads']
            memory = nworkers * self.cluster.worker_spec[0]['options']['memory_limit']
            vprint(self.verbose,
                   f'Dask cluster initiated with {nworkers} workers, ' +
                   f'{nthreads} threads, {memory/2**30:0.2f} GiB memory.')
            vprint(self.verbose, f'Dask dashboard link: {self.cluster.dashboard_link}')


    def addInVar(self,
                 name       : str,
                 dist       : rv_discrete | rv_continuous,
                 distkwargs : dict[str, Any],
                 nummap     : dict[float, Any] = None,
                 seed       : int = None,
                 datasource : Optional[str] = None,
                 ) -> None:
        """
        Add an input variable to the simulation.

        Parameters
        ----------
        name : str
            The name of this variable.
        dist : scipy.stats.rv_discrete | scipy.stats.rv_continuous
            The statistical distribution to draw from.
        distkwargs : dict
            The keyword argument pairs for the statistical distribution
            function.
        nummap : dict[float, Any], default: None
            A dictionary mapping numbers to nonnumeric values.
        seed : int
            The random seed for this variable. If None, a seed will be assigned
            based on the order added.
        datasource : str, default: None
            If the invals were imported from a file, this is the filepath. If
            generated through monaco, then None.
        """
        if name in self.vars.keys():
            raise ValueError(f"'{name}' is already a Variable")

        self.ninvars += 1
        if seed is None:
            # seed is dependent on the order added
            seed = (self.seed + self.ninvars) % 2**32
        self.invarseeds.append(seed)
        invar = InVar(name=name, dist=dist, distkwargs=distkwargs, ndraws=self.ndraws,
                      nummap=nummap, samplemethod=self.samplemethod, ninvar=self.ninvars,
                      seed=seed, firstcaseismedian=self.firstcaseismedian, autodraw=False,
                      datasource=datasource)
        self.invars[name] = invar
        self.vars[name] = invar


    def addConstVal(self,
                    name : str,
                    val  : Any,
                    ) -> None:
        """
        Add a constant value for all the cases to use.

        Parameters
        ----------
        name : str
            Name for this value.
        val : Any
            The constant value.
        """
        self.constvals[name] = val


    def setNDraws(self,
                  ndraws: int,
                  ) -> None:
        """
        Set the number of random draws to perform. Will clear the results.

        Parameters
        ----------
        ndraws : int
            The number of random draws to perform.
        """
        self.clearResults()
        self.ndraws = ndraws
        self.setFirstCaseMedian(self.firstcaseismedian)
        for invar in self.invars.values():
            invar.setNDraws(ndraws)
        if self.invars != dict():
            self.drawVars()


    def drawVars(self) -> None:
        """Draw the random values for all the input variables."""
        if self.ninvars > 0:
            vprint(self.verbose, f"Drawing random samples for {self.ninvars} input variables " +
                                 f"via the '{self.samplemethod}' method...", end=' ', flush=True)
            for invar in self.invars.values():
                if invar.datasource is None:
                    invar.draw(ninvar_max=self.ninvars)
            vprint(self.verbose, 'Done', flush=True)


    def runSim(self,
               cases : None | int | Iterable[int] = None,
               ) -> None:
        """
        Run the full simulation.

        Parameters
        ----------
        cases : None | int | Iterable[int]
            The case numbers to run. If None, then all cases are run.
        """
        cases_downselect = self.downselectCases(cases=cases)
        vprint(self.verbose, f"Running '{self.name}' Monte Carlo simulation with " +
                             f"{len(cases_downselect)}/{self.ncases} cases...", flush=True)
        self.runSimWorker(casestogenerate=cases_downselect, casestopreprocess=cases_downselect,
                          casestorun=cases_downselect, casestopostprocess=cases_downselect)


    def runIncompleteSim(self) -> None:
        """
        Run the full sim, but only the cases which previously failed to
        preprocess, run, or postprocess.
        """
        allcases = self.allCases()
        casestopreprocess  = allcases - self.casespreprocessed
        casestorun         = allcases - self.casesrun           | casestopreprocess
        casestopostprocess = allcases - self.casespostprocessed | casestopreprocess | casestorun
        casestogenerate    = casestopreprocess

        vprint(self.verbose, f"Resuming incomplete '{self.name}' Monte Carlo simulation with " +
                             f"{len(casestopostprocess)}/{self.ncases} " +
                              "cases remaining to preprocess, " +
                             f"{len(casestorun)}/{self.ncases} " +
                              "cases remaining to run, and " +
                             f"{len(casestopostprocess)}/{self.ncases} " +
                              "cases remaining to postprocess...", flush=True)

        self.runSimWorker(casestogenerate=casestogenerate,
                          casestopreprocess=casestopreprocess,
                          casestorun=casestorun,
                          casestopostprocess=casestopostprocess)


    def runSimWorker(self,
                     casestogenerate    : None | int | Iterable[int],
                     casestopreprocess  : None | int | Iterable[int],
                     casestorun         : None | int | Iterable[int],
                     casestopostprocess : None | int | Iterable[int],
                     ) -> None:
        """
        The worker function to run the full sim.

        Parameters
        ----------
        casestogenerate : None | int | Iterable[int]
            The case numbers to generate. If None, then all cases are
            generated.
        casestopreprocess : None | int | Iterable[int]
            The case numbers to preprocess. If None, then all cases are
            preprocessed.
        casestorun : None | int | Iterable[int]
            The case numbers to run. If None, then all cases are run.
        casestopostprocess : None | int | Iterable[int]
            The case numbers to postprocess. If None, then all cases are
            postprocessed.
        """
        self.starttime = datetime.now()

        if set(casestorun) in (None, self.allCases()):
            self.clearResults()  # only clear results if we are rerunning all cases

        self.runsimid = self.genID()

        if self.savesimdata or self.savecasedata:
            if not os.path.exists(self.resultsdir):
                os.makedirs(self.resultsdir)
            if self.savesimdata:
                self.saveSimToFile()

        self.drawVars()
        self.genCases(cases=casestogenerate)
        self.executeAllFcns(casestopreprocess=casestopreprocess,
                            casestorun=casestorun,
                            casestopostprocess=casestopostprocess,
                            calledfromrunsim=True)
        self.genOutVars()

        self.endtime = datetime.now()
        self.runtime = self.endtime - self.starttime

        vprint(self.verbose, f'Simulation complete! Runtime: {self.runtime}', flush=True)

        if self.savecasedata:
            self.saveCasesToFile()

        if self.savesimdata:
            self.saveSimToFile()


    def genRunSimID(self) -> None:
        """Regenerate the unique ID for this simulation run."""
        self.runsimid = self.genID()


    def genID(self) -> int:
        """
        Generate a unique ID based on the simulation seed, name, and current
        timestamp.

        Returns
        -------
        uniqueid : int
            A unique ID.
        """
        uniqueid = (self.seed + hash_str_repeatable(self.name) + hash(datetime.now())) % 2**32
        return uniqueid


    def genCases(self,
                 cases : None | int | Iterable[int] = None,
                 ) -> None:
        """
        Generate all the Monte Carlo case objects.

        Parameters
        ----------
        cases : None | int | Iterable[int]
            The case numbers to generate. If None, then all cases are
            generated.
        """
        vprint(self.verbose, 'Generating cases...', end=' ', flush=True)
        self.genCaseSeeds()

        # If we are rerunning partial cases we don't want to reset this
        if cases is None:
            self.cases = []

        cases_downselect = self.downselectCases(cases)
        for ncase in cases_downselect:
            ismedian = False
            if self.firstcaseismedian and ncase == 0:
                ismedian = True
            self.cases.append(Case(ncase=ncase, ismedian=ismedian, invars=self.invars,
                                   constvals=self.constvals, keepsiminput=self.keepsiminput,
                                   keepsimrawoutput=self.keepsimrawoutput,
                                   seed=int(self.caseseeds[ncase])))
        self.cases.sort(key=lambda case: case.ncase)
        vprint(self.verbose, 'Done', flush=True)


    def genCaseSeeds(self) -> None:
        """Generate the random seeds for each of the random cases."""
        generator = np.random.RandomState(self.seed)
        self.caseseeds = list(generator.randint(0, 2**31-1, size=self.ncases))


    def executeAllFcns(self,
                       casestopreprocess : None | int | Iterable[int] = None,
                       casestorun : None | int | Iterable[int] = None,
                       casestopostprocess : None | int | Iterable[int] = None,
                       calledfromrunsim : bool = False
                       ):
        """
        Preprocess, run, and postprocess all the Monte Carlo cases.

        Parameters
        ----------
        casestopreprocess : None | int | Iterable[int]
            The case numbers to preprocess. If None, then all cases are
            preprocessed.
        casestorun : None | int | Iterable[int]
            The case numbers to run. If None, then all cases are run.
        casestopostprocess : None | int | Iterable[int]
            The case numbers to postprocess. If None, then all cases are
            postprocessed.
        calledfromrunsim : bool, default: False
            Whether this was called from self.runSim(). If False, a new ID for
            this simulation run is generated.
        """
        if self.singlethreaded:
            self.preProcessCases(cases=casestopreprocess)
            self.runCases(cases=casestorun, calledfromrunsim=calledfromrunsim)
            self.postProcessCases(cases=casestopostprocess)
        else:
            casestopreprocess_downselect = self.downselectCases(cases=casestopreprocess)
            casestorun_downselect = self.downselectCases(cases=casestorun)
            casestopostprocess_downselect = self.downselectCases(cases=casestopostprocess)

            preprocessedcases = dict()
            runcases = dict()
            postprocessedcases = dict()
            try:
                for case in self.cases:
                    if case.ncase in casestopreprocess_downselect:
                        case.haspreprocessed = False
                    if case.ncase in casestorun_downselect:
                        case.hasrun = False
                    if case.ncase in casestopostprocess_downselect:
                        case.haspostprocessed = False

                    if case.ncase in casestopreprocess_downselect:
                        casepreprocessed_delayed = dask.delayed(preprocess_case)(
                            self.fcns[SimFunctions.PREPROCESS], case,
                            self.debug, self.verbose)
                        preprocessedcases[case.ncase] = casepreprocessed_delayed

                    if case.ncase in casestorun_downselect:
                        if case.haspreprocessed:
                            case_to_delay = case
                        else:
                            case_to_delay = preprocessedcases[case.ncase]
                        caserun_delayed = dask.delayed(run_case)(
                            self.fcns[SimFunctions.RUN], case_to_delay,
                            self.debug, self.verbose, self.runsimid)
                        runcases[case.ncase] = caserun_delayed

                    if case.ncase in casestopostprocess_downselect:
                        if case.hasrun:
                            case_to_delay = case
                        else:
                            case_to_delay = runcases[case.ncase]
                        casepostprocessed_delayed = dask.delayed(postprocess_case)(
                            self.fcns[SimFunctions.POSTPROCESS], case_to_delay,
                            self.debug, self.verbose)
                        postprocessedcases[case.ncase] = casepostprocessed_delayed


                if self.verbose:
                    x = dask.persist(*postprocessedcases.values())
                    n_tasks = (len(casestopreprocess_downselect)
                              + len(casestorun_downselect)
                              + len(casestopostprocess_downselect))
                    tqdm_dask(x, total=n_tasks,
                              desc='Preprocessing, running, and postprocessing cases',
                              unit=' cases', position=0)
                    fullyexecutedcases = dask.compute(*x)
                else:
                    fullyexecutedcases = dask.compute(*postprocessedcases.values())

            except KeyboardInterrupt:
                raise

            # Save out results
            for case in fullyexecutedcases:
                if any([case.haspreprocessed, case.hasrun, case.haspostprocessed]):
                    self.cases[case.ncase] = case
                if case.haspreprocessed:
                    self.casespreprocessed.add(case.ncase)
                if case.hasrun:
                    self.casesrun.add(case.ncase)
                if case.haspostprocessed:
                    self.casespostprocessed.add(case.ncase)


    def preProcessCases(self,
                        cases : None | int | Iterable[int] = None,
                        ) -> None:
        """
        Preprocess all the Monte Carlo cases.

        Parameters
        ----------
        cases : None | int | Iterable[int]
            The case numbers to preprocess. If None, then all cases are
            preprocessed.
        """
        cases_downselect = self.downselectCases(cases=cases)
        preprocessedcases = []

        # Single-threaded for loop
        if self.singlethreaded:
            if self.verbose:
                pbar = tqdm(total=len(cases_downselect), desc='Preprocessing cases',
                            unit=' cases', position=0)
            for case in self.cases:
                if case.ncase in cases_downselect:
                    case.haspreprocessed = False
                    case = preprocess_case(self.fcns[SimFunctions.PREPROCESS],
                                            case, self.debug, self.verbose)
                    preprocessedcases.append(case)
                    if self.verbose:
                        pbar.update(1)
            if self.verbose:
                pbar.refresh()
                pbar.close()

        # Dask parallel processing
        else:
            try:
                for case in self.cases:
                    if case.ncase in cases_downselect:
                        case.haspreprocessed = False
                        case_delayed = dask.delayed(preprocess_case)(
                            self.fcns[SimFunctions.PREPROCESS], case,
                            self.debug, self.verbose)
                        preprocessedcases.append(case_delayed)

                if self.verbose:
                    x = dask.persist(preprocessedcases)
                    tqdm_dask(x, total=len(cases_downselect),
                              desc='Preprocessing cases',
                              unit=' cases', position=0)
                    preprocessedcases = dask.compute(*x)[0]
                else:
                    preprocessedcases = dask.compute(*preprocessedcases)

            except KeyboardInterrupt:
                raise

        # Save out results
        for case in preprocessedcases:
            if case.haspreprocessed:
                self.cases[case.ncase] = case
                self.casespreprocessed.add(case.ncase)


    def runCases(self,
                 cases            : None | int | Iterable[int] = None,
                 calledfromrunsim : bool = False,
                 ) -> None:
        """
        Run all the Monte Carlo cases.

        Parameters
        ----------
        cases : None | int | Iterable[int]
            The case numbers to run. If None, then all cases are run.
        calledfromrunsim : bool, default: False
            Whether this was called from self.runSim(). If False, a new ID for
            this simulation run is generated.
        """
        cases_downselect = self.downselectCases(cases=cases)
        runcases = []

        if not calledfromrunsim:
            self.runsimid = self.genID()

        # Single-threaded for loop
        if self.singlethreaded:
            if self.verbose:
                pbar = tqdm(total=len(cases_downselect), desc='Running cases',
                            unit=' cases', position=0)
            for case in self.cases:
                if case.ncase in cases_downselect:
                    case.hasrun = False
                    case = run_case(self.fcns[SimFunctions.RUN], case,
                                    self.debug, self.verbose, self.runsimid)
                    runcases.append(case)
                    if self.verbose:
                        pbar.update(1)
            if self.verbose:
                pbar.refresh()
                pbar.close()

        # Dask parallel processing
        else:
            try:
                for case in self.cases:
                    if case.ncase in cases_downselect:
                        case.hasrun = False
                        case_delayed = dask.delayed(run_case)(
                            self.fcns[SimFunctions.RUN], case,
                            self.debug, self.verbose, self.runsimid)
                        runcases.append(case_delayed)

                if self.verbose:
                    x = dask.persist(runcases)
                    tqdm_dask(x, total=len(cases_downselect),
                              desc='Running cases',
                              unit=' cases', position=0)
                    runcases = dask.compute(*x)[0]
                else:
                    runcases = dask.compute(*runcases)

            except KeyboardInterrupt:
                raise

        # Save out results
        for case in runcases:
            if case.hasrun:
                self.cases[case.ncase] = case
                self.casesrun.add(case.ncase)


    def postProcessCases(self,
                         cases : None | int | Iterable[int] = None,
                         ) -> None:
        """
        Postprocess all the Monte Carlo cases.

        Parameters
        ----------
        cases : None | int | Iterable[int]
            The case numbers to postprocess. If None, then all cases are
            postprocessed.
        """
        cases_downselect = self.downselectCases(cases=cases)
        postprocessedcases = []

        # Single-threaded for loop
        if self.singlethreaded:
            if self.verbose:
                pbar = tqdm(total=len(cases_downselect), desc='Postprocessing cases',
                            unit=' cases', position=0)
            for case in self.cases:
                if case.ncase in cases_downselect:
                    case.haspostprocessed = False
                    case = postprocess_case(self.fcns[SimFunctions.POSTPROCESS],
                                            case, self.debug, self.verbose)
                    postprocessedcases.append(case)
                    if self.verbose:
                        pbar.update(1)
            if self.verbose:
                pbar.refresh()
                pbar.close()

        # Dask parallel processing
        else:
            try:
                for case in self.cases:
                    if case.ncase in cases_downselect:
                        case.haspostprocessed = False
                        case_delayed = dask.delayed(postprocess_case)(
                            self.fcns[SimFunctions.POSTPROCESS], case,
                            self.debug, self.verbose)
                        postprocessedcases.append(case_delayed)

                if self.verbose:
                    x = dask.persist(postprocessedcases)
                    tqdm_dask(x, total=len(cases_downselect),
                              desc='Postprocessing cases',
                              unit=' cases', position=0)
                    postprocessedcases = dask.compute(*x)[0]
                else:
                    postprocessedcases = dask.compute(*postprocessedcases)

            except KeyboardInterrupt:
                raise

        # Save out results
        for case in postprocessedcases:
            if case.haspostprocessed:
                self.cases[case.ncase] = case
                self.casespostprocessed.add(case.ncase)


    def genOutVars(self,
                   datasource : Optional[str] = None,
                   ) -> None:
        """
        Generate the output variables.

        Parameters
        ----------
        datasource : str, default: None
            If the outvals were imported from a file, this is the filepath. If
            generated through monaco, then None.
        """
        for i_var, varname in enumerate(self.cases[0].outvals.keys()):
            if varname in self.invars.keys():
                raise ValueError(f"'{varname}' is already a Variable")

            # seed is dependent on the order added, used for bootstrapping
            seed = (self.seed - 1 - i_var) % 2**32
            self.outvarseeds.append(seed)

            vals = []
            for i in range(self.ncases):
                vals.append(self.cases[i].outvals[varname].val)

            if self.cases[0].outvals[varname].valmapsource == 'auto':
                uniquevals : set[Any] = set()
                valmap : dict[Any, float] = None
                for i in range(self.ncases):
                    if self.cases[i].outvals[varname].valmap is None:
                        uniquevals = None
                    else:
                        uniquevals.update(self.cases[i].outvals[varname].valmap.keys())
                if uniquevals is not None:
                    valmap = dict()
                    for i, val in enumerate(uniquevals):
                        valmap[val] = i
            else:
                valmap = self.cases[0].outvals[varname].valmap

            outvar = OutVar(name=varname, vals=vals, valmap=valmap,
                            ndraws=self.ndraws, seed=seed,
                            firstcaseismedian=self.firstcaseismedian,
                            datasource=datasource)
            self.outvars[varname] = outvar
            self.vars[varname] = outvar
            for i in range(self.ncases):
                self.cases[i].outvars[varname] = outvar

        self.noutvars = len(self.outvars)


    def scalarOutVars(self) -> dict[str, OutVar]:
        """
        Return a dict of just the scalar output variables.
        """
        scalaroutvars = dict()
        if self.outvars != dict():
            for name, outvar in self.outvars.items():
                if outvar.isscalar:
                    scalaroutvars[name] = outvar

        return scalaroutvars


    def calcSensitivities(self,
                          outvarnames : None | str | Iterable[str] = None,
                          tol         : float = 1e-6,
                          verbose     : bool = False,
                          ) -> None:
        """
        Calculate the sensitivity indices for the specified outvars.

        Parameters
        ----------
        outvarnames : None | str | Iterable[str] (default: None)
            The outvar names to calculate sensitivity indices for. If None,
            then calculates sensitivities for all scalar outvars.
        tol : float, default 1e-6
            The convergence tolerance for scipy's minimize function acting on the
            negative log likelihood function.
        verbose : bool, default False
            Whether to print diagnostic information.
        """
        if outvarnames is None:
            outvarnames = list(self.scalarOutVars().keys())
        outvarnames = get_list(outvarnames)

        for outvarname in outvarnames:
            if not self.outvars[outvarname].isscalar:
                vwarn(self.verbose, f"Output variable '{outvarname}' is not scalar," +
                                     'skipping sensitivity calculations.')
            else:
                vprint(self.verbose, f"Calculating sensitivity indices for '{outvarname}'...")
                sensitivities, ratios = calc_sensitivities(self, outvarname,
                                                           tol=tol, verbose=verbose)

                sensitivities_dict = dict()
                ratios_dict = dict()
                for i, name in enumerate(self.invars.keys()):
                    sensitivities_dict[name] = sensitivities[i]
                    ratios_dict[name] = ratios[i]

                self.outvars[outvarname].sensitivity_indices = sensitivities_dict
                self.outvars[outvarname].sensitivity_ratios = ratios_dict
                vprint(self.verbose, "Done calculating sensitivity indices.")


    def genCovarianceMatrix(self) -> None:
        """
        Generate the covariance matrix and correlation coefficients between all
        the scalar variables.
        """
        self.covvarlist = []
        allnums = []
        for var in self.vars.keys():
            if self.vars[var].isscalar:
                allnums.append(self.vars[var].nums)
                self.covvarlist.append(self.vars[var].name)
        self.covs = np.cov(np.array(allnums))
        self.corrcoeffs = np.corrcoef(np.array(allnums))

        for i, coeff in enumerate(self.corrcoeffs[0]):
            if np.isnan(coeff):
                vwarn(self.verbose, "Unable to generate correlation coefficient for " +
                                   f"'{self.covvarlist[i]}'. This may happen if this variable " +
                                    "does not vary, or if an infinite value was drawn.")


    def corr(self) -> tuple[np.ndarray, list[str]]:
        """
        Generate a correlation matrix between all the scalar variables.

        Returns
        -------
        (corcoeffs, covvarlist) : (numpy.ndarray, list[str])
            corrcoeffs is a correlation matrix between all the scalar input
            and output variables.
            covvarlist is a list of all the scalar input and output variables.
        """
        self.genCovarianceMatrix()
        return self.corrcoeffs, self.covvarlist


    def cov(self) -> tuple[np.ndarray, list[str]]:
        """
        Generate a covariance matrix between all the scalar variables.

        Returns
        -------
        (covs, covvarlist) : (numpy.ndarray, list[str])
            covs is a covariance matrix between all the scalar input and
            output variables.
            covvarlist is a list of all the scalar input and output variables.
        """
        self.genCovarianceMatrix()
        return self.covs, self.covvarlist


    def plot(self,
             scalarvars  : Optional[list[InVar | OutVar | str]] = None,
             cases           : None | int | Iterable[int] = None,
             highlight_cases : None | int | Iterable[int] = empty_list(),
             rug_plot    : bool   = False,
             cov_plot    : bool   = False,
             cov_p       : None | float | Iterable[float] = None,
             invar_space : InVarSpace | Iterable[InVarSpace] = InVarSpace.NUMS,
             fig         : Figure = None,
             title       : str    = '',
             plotkwargs  : dict   = dict(),
             ) -> tuple[Figure, tuple[Axes, ...]]:
        """
        Plot all the scalar variables against each other in a grid.

        Parameters
        ----------
        scalarvars : list[monaco.mc_var.InVar | monaco.mc_var.OutVar | str]
            The variables to plot. If None, then grabs all the input variables
            and scalar output variables.
        cases : None | int | Iterable[int], default: None
            The cases to plot. If None, then all cases are plotted.
        highlight_cases : None | int | Iterable[int], default: []
            The cases to highlight. If [], then no cases are highlighted.
        rug_plot : bool, default: False
            Whether to plot rug marks.
        cov_plot : bool, default: False
            Whether to plot a covariance ellipse at a certain gaussian percentile
            level.
        cov_p : None | float | Iterable[float], default: None
            The gaussian percentiles for the covariance plot.
        invar_space : monaco.InVarSpace | Iterable[InVarSpace], default: 'nums'
            The space to plot invars in, either 'nums' or 'pcts'. If an iterable,
            specifies this individually for each of varx, vary, and varz.
        fig : matplotlib.figure.Figure, default: None
            The figure handle to plot in. If None, a new figure is created.
        title : str, default: ''
            The figure title.

        Returns
        -------
        (fig, axes) : (matplotlib.figure.Figure, (matplotlib.axes.Axes, ...))
            fig is the figure handle for the plot.
            axes is a tuple of the axes handles for the plots.
        """
        if scalarvars is None:
            scalarvars = []
            for invar in self.invars.values():
                scalarvars.append(invar)
            for scalaroutvar in self.scalarOutVars().values():
                scalarvars.append(scalaroutvar)

        scalarvars = get_list(scalarvars)

        for i, scalarvar in enumerate(scalarvars):
            if isinstance(scalarvar, str):
                scalarvars[i] = self.vars[scalarvar]

        invars = []
        scalaroutvars = []
        for scalarvar in scalarvars:
            if isinstance(scalarvar, InVar):
                invars.append(scalarvar)
            elif isinstance(scalarvar, OutVar) and scalarvar.isscalar:
                scalaroutvars.append(scalarvar)

        fig, axs = multi_plot_grid_rect(varsx=invars,
                                        varsy=scalaroutvars,
                                        cases=cases, highlight_cases=highlight_cases,
                                        rug_plot=rug_plot,
                                        cov_plot=cov_plot, cov_p=cov_p,
                                        invar_space=invar_space,
                                        fig=fig, title=title, plotkwargs=plotkwargs)
        return fig, axs


    def clearResults(self) -> None:
        """Clear all the simulation results."""
        for varname in self.outvars.keys():
            if varname in self.vars.keys():
                del self.vars[varname]
        self.cases = []
        self.outvars = dict()
        self.casespreprocessed = set()
        self.casesrun = set()
        self.casespostprocessed = set()
        self.corrcoeff = None
        self.covcoeff = None
        self.covvarlist = None
        self.endtime = None
        self.runtime = None
        self.runsimid = self.genID()


    def reset(self) -> None:
        """Completely reset the simulation to the default object state."""
        self.clearResults()
        self.vars = dict()
        self.invars = dict()
        self.constvals = dict()
        self.ninvars = 0
        self.noutvars = 0
        self.invarseeds = []
        self.caseseeds = []
        self.starttime = None


    def downselectCases(self,
                        cases : None | int | Iterable[int] = None,
                        ) -> set[int]:
        """
        Convert the `cases` input to a set of all the target cases.

        Parameters
        ----------
        cases : None | int | Iterable[int]
            The case numbers to downselect. If None, returns all cases.

        Returns
        -------
        cases_downselect : set[int]
            A set of all the case numbers to use.
        """
        if cases is None:
            cases_downselect = self.allCases()
        else:
            cases_downselect = set(get_list(cases))
        return cases_downselect


    def allCases(self) -> set[int]:
        """
        Get a set of the indices for all the cases.

        Returns
        -------
        allCases : set[int]
            A set of all the case numbers.
        """
        allCases = set(range(self.ncases))
        return allCases


    def exportVars(self,
                   vars : dict[str, InVar | OutVar],
                   filename : Optional[str | pathlib.Path],
                   ) -> pathlib.Path:
        """
        Export the nums for the selected to file for use externally.

        Parameters
        ----------
        vars : dict[str, InVar | OutVar]
            The vars to save to file. Will export the nums, not the vals.
        filename : Optional[str | pathlib.Path]
            The file to save to. Must be a csv or json.
            If a str, then will save in the resultsdir.

        Returns
        -------
        filepath : pathlib.Path
            The filepath the vars were saved to.
        """
        if vars == []:
            raise ValueError('No vars to save to file.')

        if isinstance(filename, str):
            filepath = self.resultsdir / filename
        elif isinstance(filename, pathlib.Path):
            filepath = filename

        if filepath.suffix.lower() not in ('.csv', '.json'):
            raise ValueError(f"'{filename}' must be a .csv or .json file.")
        if filepath.exists():
            vwarn(self.verbose, f'{filepath.name} already exists, overwriting.')

        if filepath.suffix.lower() == '.csv':
            varnames = list(vars.keys())
            for i, var in enumerate(vars.values()):
                if i == 0:
                    data_csv = np.array(var.nums)
                else:
                    data_csv = np.vstack([data_csv, np.array(var.nums)])

            with open(filepath, 'w', newline='') as f:
                writer = csv.writer(f, quoting=csv.QUOTE_NONNUMERIC)
                writer.writerow(varnames)
                for i in range(self.ncases):
                    if len(data_csv.shape) == 1:
                        writer.writerow((data_csv[i], ))
                    else:
                        writer.writerow(data_csv[:, i])

        elif filepath.suffix.lower() == '.json':
            data_json = dict()
            for varname, var in vars.items():
                data_json[varname] = np.array(var.nums).tolist()

            with open(filepath, 'w', newline='') as f:
                json.dump(data_json, f, indent=0)

        return filepath


    def exportInVars(self,
                     filename : Optional[str | pathlib.Path] = None,
                     ) -> None:
        """
        Export the drawn nums for all the invars to file for use externally.

        Parameters
        ----------
        filename : Optional[str | pathlib.Path]
            The file to save to. Must be a csv or json.
            If a str, then will save in the resultsdir.
            If None, then will save to '{self.name}_invarnums.json'.
        """
        vprint(self.verbose, 'Exporting InVar draws to file...', flush=True)

        if filename is None:
            filename = self.resultsdir / f'{self.name}_invarnums.json'

        filepath = self.exportVars(self.invars, filename)

        vprint(self.verbose, f"InVar nums saved in '{filepath.name}'", flush=True)


    def exportOutVars(self,
                      filename : Optional[str | pathlib.Path] = None,
                      ) -> None:
        """
        Export the nums for all the outvars to file for use externally.

        Parameters
        ----------
        filename : Optional[str | pathlib.Path]
            The file to save to. Must be a csv or json.
            If a str, then will save in the resultsdir.
            If None, then will save to '{self.name}_outvarnums.json'.
        """
        vprint(self.verbose, 'Exporting InVar draws to file...', flush=True)

        if filename is None:
            filename = self.resultsdir / f'{self.name}_outvarnums.json'

        filepath = self.exportVars(self.outvars, filename)

        vprint(self.verbose, f"OutVar nums saved in '{filepath.name}'", flush=True)


    def importVars(self,
                   filepath : str | pathlib.Path,
                   ) -> tuple[dict[str, list[Any]], pathlib.Path]:
        """
        Import values from an external file.

        Parameters
        ----------
        filepath : str | pathlib.Path
            The file to load from. Must be a csv or json.

        Returns
        -------
        data : dict[str, Any]
            A dictionary where the keys are the variable name and the values
            are a list of the vals.
        filepath : pathlib.Path
            The full filepath for the file which was loaded from.
        """
        if isinstance(filepath, str):
            filepath = pathlib.Path(filepath)

        if filepath.suffix.lower() not in ('.csv', '.json'):
            raise ValueError(f"'{filepath.name}' must be a .csv or .json file.")

        if self.cases == []:
            self.genCases()

        data : dict[str, list] = dict()
        if filepath.suffix.lower() == '.csv':
            with open(filepath, 'r') as f:
                reader = csv.reader(f, quoting=csv.QUOTE_NONNUMERIC)
                headers = next(reader)
                for outvalname in headers:
                    data[outvalname] = []
                for row in reader:
                    for outvalname, vals in zip(headers, row):
                        data[outvalname].append(vals)

        elif filepath.suffix.lower() == '.json':
            with open(filepath, 'r') as f:
                data = json.load(f)

        for vals_list in data.values():
            if len(vals_list) != self.ncases:
                raise ValueError(f'Length of data ({len(vals_list)}) must match ' +
                                 f'the number of sim cases ({self.ncases}).')

        return data, filepath


    def importInVars(self,
                     filepath     : str | pathlib.Path,
                     dists        : Optional[list[rv_discrete | rv_continuous]] = None,
                     distskwargs  : Optional[list[dict[str, Any]]] = None,
                     nummaps      : Optional[list[dict[Any, float]]] = None,
                     ) -> None:
        """
        Import draws from an external file as InVals. For each of the keyword
        arguments, they must be the same length as the number of invars.

        Parameters
        ----------
        filepath : str | pathlib.Path
            The file to load from. Must be a csv or json.
        dists : list[rv_discrete | rv_continuous], default: None
            A list of the distribution that was used for the draws. Needed if
            it is desired to plot the analytical distribution or run a DVARS
            sensitivity analysis.
        distskwargs : list[dict[str, Any]], default: None
            A list of the distribution kwargs that were used for the draws.
            Needed if it is desired to plot the analytical distribution or run
            a DVARS sensitivity analysis.
        nummaps : list[dict[Any, float]], default: None
            A list of nummap dicts mapping numbers to nonnumeric values.
        """
        vprint(self.verbose, 'Importing InVals from file...', flush=True)

        data, filepath = self.importVars(filepath)

        if dists is not None and distskwargs is not None:
            dists = get_list(dists)
            distskwargs = get_list(distskwargs)
            if len(dists) != len(data):
                raise ValueError(f'Length of dists ({len(dists)}) ',
                                 f'must match the number of invars ({len(data)}).')
            if len(distskwargs) != len(data):
                raise ValueError(f'Length of distskwargs ({len(distskwargs)}) ',
                                 f'must match the number of invars ({len(data)}).')
        else:
            dists = [None for _ in range(len(data))]
            distskwargs = [None for _ in range(len(data))]

        if nummaps is not None:
            nummaps = get_list(nummaps)
            if len(nummaps) != len(data):
                raise ValueError(f'Length of nummaps ({len(nummaps)}) ',
                                 f'must match the number of invars ({len(data)}).')
        else:
            nummaps = [None for _ in range(len(data))]

        for i, (valname, nums) in enumerate(data.items()):

            if dists[i] is None or distskwargs[i] is None:
                pcts = [None for _ in range(self.ncases)]
                if distskwargs[i] is None:
                    distskwargs[i] = dict()
            else:
                dist = dists[i](**distskwargs[i])
                pcts = np.array(dist.cdf(nums))

            self.addInVar(name=valname, dist=dists[i], distkwargs=distskwargs[i], nummap=nummaps[i],
                          seed=None, datasource=str(filepath.resolve()))
            nums = [np.array(num) for num in nums]
            self.invars[valname].nums = nums
            self.invars[valname].pcts = pcts
            self.invars[valname].mapNums()

            for case in self.cases:
                case.invars[valname] = self.invars[valname]
                case.invals[valname] = self.invars[valname].getVal(case.ncase)


        vprint(self.verbose, f"InVals loaded from '{filepath.name}' and converted to variables",
               flush=True)


    def importOutVars(self,
                      filepath : str | pathlib.Path,
                      nummaps  : Optional[list[dict[float, Any]]] = None,
                      ) -> None:
        """
        Import results from an external file as OutVals, convert to OutVars.

        Parameters
        ----------
        filepath : str | pathlib.Path
            The file to load from. Must be a csv or json.
        nummaps : list[dict[float, Any]], default: None
            A list of nummap dicts mapping numbers to nonnumeric values. Note
            that this is reversed from providing valmaps to OutVals.
        """
        vprint(self.verbose, 'Importing OutVals from file...', flush=True)

        data, filepath = self.importVars(filepath)

        if nummaps is not None:
            nummaps = get_list(nummaps)
            if len(nummaps) != len(data):
                raise ValueError(f'Length of nummaps ({len(nummaps)}) ',
                                 f'must match the number of outvars ({len(data)}).')
            valmaps = []
            for nummap in nummaps:
                if nummap is not None:
                    valmaps.append({val: num for num, val in nummap.items()})
                else:
                    valmaps.append(None)
        else:
            nummaps = [None for _ in range(len(data))]
            valmaps = [None for _ in range(len(data))]

        for i, (valname, nums) in enumerate(data.items()):
            if nummaps[i] is None:
                vals = nums
            else:
                vals = [nummaps[i][num] for num in nums]
            for case in self.cases:
                case.addOutVal(valname, vals[case.ncase], valmap=valmaps[i])

        self.genOutVars(datasource=str(filepath.resolve()))

        vprint(self.verbose, f"OutVals loaded from '{filepath.name}' and converted to variables",
               flush=True)


    def saveSimToFile(self) -> None:
        """Save the simulation to a .mcsim file"""
        if self.savesimdata:
            vprint(self.verbose, 'Saving sim results to file...', flush=True)

            try:
                self.filepath.unlink()
            except FileNotFoundError:
                pass
            self.filepath.touch()
            with open(self.filepath, 'wb') as file:
                cloudpickle.dump(self, file)

            vprint(self.verbose, f"Sim results saved in '{self.filepath}'", flush=True)


    def saveCasesToFile(self,
                        cases : None | int | Iterable[int] = None,
                        ) -> None:
        """
        Save the specified cases to .mccase files.

        Parameters
        ----------
        cases : None | int | Iterable[int]
            The cases to save. If None, save all cases.
        """
        cases_downselect = self.downselectCases(cases=cases)

        if self.savecasedata:
            vprint(self.verbose, 'Saving cases to file...', flush=True)

            for ncase in cases_downselect:
                filepath = self.resultsdir / f'{self.name}_{ncase}.mccase'
                self.cases[ncase].filepath = filepath
                try:
                    filepath.unlink()
                except FileNotFoundError:
                    pass
                with open(filepath, 'wb') as file:
                    cloudpickle.dump(self.cases[ncase], file)

            vprint(self.verbose, f"Raw case results saved in '{self.resultsdir}'",
                   flush=True)


    def loadCases(self) -> None:
        """Load the data for cases from file."""
        vprint(self.verbose, f'{self.filepath} indicates {len(self.casesrun)}/{self.ncases} ' +
                              'cases were run, attempting to load raw case data from disk...',
                             end='\n', flush=True)
        self.cases = []
        casesloaded = set()
        casesstale  = set()
        casesnotloaded        = self.allCases()
        casesnotpostprocessed = self.allCases()

        # pbar = tqdm(total=len(self.casesrun), unit=' cases', desc='Loading', position=0)

        for ncase in self.casesrun:
            filepath = self.resultsdir / f'{self.name}_{ncase}.mccase'
            try:
                with open(filepath, 'rb') as file:
                    try:
                        case = cloudpickle.load(file)
                        if (not case.haspreprocessed) \
                            or (not case.hasrun) \
                            or (case.runtime is None):  # only load case if it completed running
                            self.cases.append(None)
                            vwarn(self.verbose, f'{filepath.name} did not finish running, ' +
                                                 'not loaded')
                        else:
                            self.cases.append(case)

                            if case.runsimid != self.runsimid:
                                vwarn(self.verbose, f'{filepath.name} is not from the most ' +
                                                     'recent run and may be stale')
                                casesstale.add(ncase)
                            casesloaded.add(ncase)
                            casesnotloaded.remove(ncase)
                            if ncase in self.casespostprocessed:
                                casesnotpostprocessed.remove(ncase)

                    except Exception:
                        vwarn(self.verbose, f'Unknown error loading {filepath.name}')

            except FileNotFoundError:
                vwarn(self.verbose, f'{filepath.name} expected but not found')
            # pbar.update(1)

        self.casespreprocessed  = set(casesloaded)
        self.casesrun           = set(casesloaded)
        self.casespostprocessed = set(casesloaded) - casesnotpostprocessed
        # pbar.refresh()
        # pbar.close()
        vprint(self.verbose, f'\nData for {len(casesloaded)}/{self.ncases} cases loaded from disk',
               flush=True)

        if casesnotloaded != set():
            vwarn(self.verbose, 'The following cases were not loaded: ' +
                               f'[{", ".join([str(i) for i in sorted(casesnotloaded)])}]')
        if casesnotpostprocessed != set():
            vwarn(self.verbose, 'The following cases have not been postprocessed: ' +
                               f'[{", ".join([str(i) for i in sorted(casesnotpostprocessed)])}]')
        if casesstale != set():
            vwarn(self.verbose, 'The following cases were loaded but may be stale: ' +
                               f'[{", ".join([str(i) for i in sorted(casesstale)])}]')

        extrafiles = self.findExtraResultsFiles()
        if extrafiles != set():
            vwarn(self.verbose, 'The following extra .mcsim and .mccase files were found in the ' +
                                'results directory, run removeExtraResultsFiles() to clean them ' +
                               f'up: [{", ".join([str(i) for i in sorted(extrafiles)])}]')


    def findExtraResultsFiles(self) -> set[str]:
        """
        Find .mcsim and .mccase files that we don't expect to see in the
        results directory.

        Returns
        -------
        filenames : set[str]
            The extra files.
        """
        files = set(self.resultsdir.glob('**/*.mcsim')) | set(self.resultsdir.glob('**/*.mccase'))
        filenames = set(file.name for file in files)
        try:
            filenames.remove(f'{self.name}.mcsim')
        except Exception:
            pass
        for ncase in range(self.ncases):
            try:
                filenames.remove(f'{self.name}_{ncase}.mccase')
            except Exception:
                pass

        return filenames


    def removeExtraResultsFiles(self) -> None:
        """
        Delete all unexpected .mcsim and .mccase files in the results
        directory.
        """
        extrafiles = self.findExtraResultsFiles()
        for file in extrafiles:
            filepath = self.resultsdir / file
            filepath.unlink()



# mc_var.py
from __future__ import annotations

import numpy as np
from scipy.stats import rv_continuous, rv_discrete, describe
from scipy.stats.stats import DescribeResult
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from monaco.mc_val import Val, InVal, OutVal
from monaco.mc_varstat import VarStat
from monaco.mc_enums import SampleMethod, Sensitivities, VarStatType, InVarSpace
from monaco.mc_sampling import sampling
from monaco.mc_plot import plot, plot_sensitivities
from monaco.helper_functions import empty_list, hashable_val
from copy import copy
from typing import Any, Callable, Iterable, Optional
from warnings import warn
from abc import ABC, abstractmethod


### Var Base Class ###
class Var(ABC):
    """
    Abstract base class to hold the data for a Monte Carlo variable.

    Parameters
    ----------
    name : str
        The name of this variable.
    ndraws : int
        The number of random draws.
    seed : int
        The random seed to use for bootstrapping.
    firstcaseismedian : bool
        Whether the first case represents the median case.
    datasource : str
        If the vals were imported from a file, this is the filepath. If
        generated through monaco, then None.

    Attributes
    ----------
    name : str
        The name of this variable.
    ndraws : int
        The number of random draws.
    seed : int
        The random seed to use for bootstrapping.
    firstcaseismedian : bool
        Whether the first case represents the median case.
    datasource : str
        If the vals were imported from a file, this is the filepath. If
        generated through monaco, then None.
    ncases : int
        The number of cases, which is `ndraws + 1` if `firstcaseismedian` and
        `ndraws` otherwise.
    vals : list[Any]
        The values corresponding to the randomly drawn numbers. If valmap is
        None, then `vals == nums.tolist()`.
    valmap : dict[Any, float]
        A dictionary mapping nonnumeric values to numbers (the inverse of
        `nummap`).
    nums : list[np.ndarray]
        The randomly drawn numbers obtained by feeding `pcts` into `dist`.
    nummap : dict[float, Any]
        A dictionary mapping numbers to nonnumeric values (the inverse of
        `valmap`).
    pcts : list[float]
        The randomly drawn percentiles.
    maxdim : int
        The maximum dimension of the values.
    isscalar : bool
        Whether this is a scalar variable.
    varstats : list[moncao.mc_varstat.VarStat]
        A list of all the variable statistics for this variable.
    """
    def __init__(self,
                 name              : str,
                 ndraws            : int,
                 seed              : int,
                 firstcaseismedian : bool,
                 datasource        : Optional[str],
                 ):
        self.name = name
        self.ndraws = ndraws
        self.seed = seed
        self.firstcaseismedian = firstcaseismedian
        self.datasource = datasource

        self.ncases = ndraws + 1
        self.setFirstCaseMedian(firstcaseismedian)
        self.vals     : list[Any]
        self.valmap   : dict[Any, float]
        self.nums     : list[np.ndarray]
        self.nummap   : dict[float, Any]
        self.pcts     : list[float]
        self.maxdim   : int
        self.isscalar : bool
        self.varstats : list[VarStat] = empty_list()


    def setFirstCaseMedian(self,
                           firstcaseismedian : bool,
                           ) -> None:
        """
        Generate `val` based on the drawn number and the nummap.

        Parameters
        ----------
        firstcaseismedian : bool
            Whether the first case represents a median case.
        """
        if firstcaseismedian:
            self.firstcaseismedian = True
            self.ncases = self.ndraws + 1
        else:
            self.firstcaseismedian = False
            self.ncases = self.ndraws


    def stats(self) -> DescribeResult:
        """
        Calculates statistics of the variable nums from scipy.stats.describe.

        Returns
        -------
        stats : DescribeResult
            A dict with descriptive statistics for the variable nums.
        """
        stats = describe(self.nums)
        return stats


    def addVarStat(self,
                   stat        : VarStatType | Callable,
                   statkwargs  : dict[str, Any] | None = None,
                   bootstrap   : bool = False,
                   bootstrap_k : int = 10,
                   conf        : float = 0.95,
                   seed        : int | None = None,
                   name        : str | None = None,
                   ) -> None:
        """
        Add a variable statistic to this variable.

        Parameters
        ----------
        stat : monaco.mc_enums.VarStatType | Callable
            The type of variable statistic to add.
        statkwargs : dict[str, Any]
            Keyword arguments for the specified variable stastistic.
        seed : int
            The random seed to use for bootstrapping.
        name : str
            The name of the variable statistic to add.
        """
        if statkwargs is None:
            statkwargs = dict()
        if seed is None:
            # seed is dependent on the order added
            seed = (self.seed + 1 + len(self.varstats)) % 2**32

        self.varstats.append(VarStat(var=self, stat=stat, statkwargs=statkwargs,
                                     bootstrap=bootstrap, bootstrap_k=bootstrap_k,
                                     conf=conf, seed=seed, name=name))


    def clearVarStats(self) -> None:
        """
        Remove all the variable statistics for this variable.
        """
        self.varstats = []


    def plot(self,
             vary   : InVar | OutVar | None = None,
             varz   : InVar | OutVar | None = None,
             cases           : None | int | Iterable[int] = None,
             highlight_cases : None | int | Iterable[int] = empty_list(),
             rug_plot    : bool           = False,
             cov_plot    : bool           = False,
             cov_p       : None | float | Iterable[float] = None,
             invar_space : InVarSpace | Iterable[InVarSpace] = InVarSpace.NUMS,
             ax          : Optional[Axes] = None,
             title       : str            = '',
             ) -> tuple[Figure, Axes]:
        """
        Plot this variable, against other variables if desired.
        See monaco.mc_plot.plot() for API details.
        """
        fig, ax = plot(varx=self, vary=vary, varz=varz,
                       cases=cases, highlight_cases=highlight_cases,
                       rug_plot=rug_plot, cov_plot=cov_plot, cov_p=cov_p,
                       invar_space=invar_space,
                       ax=ax, title=title)
        return fig, ax


    @abstractmethod
    def getVal(self,
               ncase : int,
               ) -> Val:
        pass



### InVar Class ###
class InVar(Var):
    """
    A Monte Carlo input variable.

    Parameters
    ----------
    name : str
        The name of this variable.
    ndraws : int
        The number of random draws.
    dist : scipy.stats.rv_discrete | scipy.stats.rv_continuous
        The statistical distribution to draw from.
    distkwargs : dict
        The keyword argument pairs for the statistical distribution function.
    nummap : dict[float, Any], default: None
        A dictionary mapping numbers to nonnumeric values (the inverse of
        `valmap`).
    samplemethod : monaco.mc_enums.SampleMethod, default: 'sobol_random'
        The random sampling method to use.
    ninvar : int
        The number of the input variable this is.
    seed : int, default: np.random.get_state(legacy=False)['state']['key'][0]
        The random seed for drawing this variable and bootstrapping.
    firstcaseismedian : bool, default: False
        Whether the first case represents the median case.
    autodraw : bool, default: True
        Whether to draw the random values when this variable is created.
    datasource : str, default: None
        If the invals were imported from a file, this is the filepath. If
        generated through monaco, then None.

    Attributes
    ----------
    isscalar : bool
        Whether this is a scalar variable. Alway True for an input variable.
    maxdim : int
        The maximum dimensions of the values. Always 0 for an input variable.
    valmap : dict[Any, float]
        A dictionary mapping nonnumeric values to numbers (the inverse of
        `nummap`).
    pcts : list[float]
        The randomly drawn percentiles.
    nums : list[np.ndarry]
        The randomly drawn numbers obtained by feeding `pcts` into `dist`.
    vals : list[Any]
        The values corresponding to the randomly drawn numbers. If valmap is
        None, then `vals == nums.tolist()`
    varstats : list[moncao.mc_varstat.VarStat]
        A list of all the variable statistics for this variable.
    """
    def __init__(self,
                 name              : str,
                 ndraws            : int,
                 dist              : rv_discrete | rv_continuous,
                 distkwargs        : dict             = None,
                 nummap            : dict[float, Any] = None,
                 samplemethod      : SampleMethod     = SampleMethod.SOBOL_RANDOM,
                 ninvar            : int              = None,
                 seed              : int = np.random.get_state(legacy=False)['state']['key'][0],
                 firstcaseismedian : bool             = False,
                 autodraw          : bool             = True,
                 datasource        : Optional[str]    = None,
                 ):
        super().__init__(name=name, ndraws=ndraws, seed=seed,
                         firstcaseismedian=firstcaseismedian,
                         datasource=datasource)

        self.dist = dist
        if distkwargs is None:
            distkwargs = dict()
        self.distkwargs = distkwargs
        self.samplemethod = samplemethod
        self.ninvar = ninvar
        self.nummap = None
        if nummap is not None:
            self.nummap = nummap

        self.isscalar = True
        self.maxdim = 0

        self.genValMap()
        if autodraw:
            self.draw(ninvar_max=None)


    def mapNums(self) -> None:
        """
        Generate `vals` based on the drawn numbers and the nummap.
        """
        self.vals = copy(self.nums)
        if self.nummap is not None:
            for i in range(self.ncases):
                self.vals[i] = self.nummap[self.nums[i].item()]


    def genValMap(self) -> None:
        """
        Generate `valmap` by inverting `nummap`.
        """
        if self.nummap is None:
            self.valmap = None
        else:
            self.valmap = {hashable_val(val): num for num, val in self.nummap.items()}


    def setNDraws(self,
                  ndraws : int,
                  ) -> None:
        """
        Set the number of random draws.

        Parameters
        ----------
        ndraws : int
            The number of random input draws.
        """
        self.ndraws = ndraws
        self.setFirstCaseMedian(self.firstcaseismedian)


    def draw(self,
             ninvar_max : int = None,
             ) -> None:
        """
        Perform the random draws based on the sampling method and the
        statistical distribution, and map those draws to values.

        Parameters
        ----------
        ninvar_max : int
            The total number of input variables for the simulation.
        """
        self.pcts = []
        self.nums = []
        dist = self.dist(**self.distkwargs)

        if self.firstcaseismedian:
            self.ncases = self.ndraws + 1
            self.pcts.append(0.5)
            self.nums.append(np.array(self.getDistMedian()))

        pcts = sampling(ndraws=self.ndraws, method=self.samplemethod,
                        ninvar=self.ninvar, ninvar_max=ninvar_max,
                        seed=self.seed)
        self.pcts.extend(pcts)
        self.nums.extend([np.array(x) for x in dist.ppf(pcts)])

        if any(np.isinf(self.nums)):
            warn( 'Infinite value drawn. Check distribution and parameters: ' +
                 f'{self.dist=}, {self.distkwargs=}')
            if self.samplemethod in (SampleMethod.SOBOL, SampleMethod.HALTON):
                warn(f'Infinite value draw may happen with {self.dist=} for the ' +
                     f'first point of the {self.samplemethod} sampling method. ' +
                     f'Consider using {SampleMethod.SOBOL_RANDOM} instead.')

        if any(np.isnan(num) for num in self.nums):
            raise ValueError( 'Invalid draw. Check distribution and parameters: ' +
                             f'{self.dist=}, {self.distkwargs=}')

        self.mapNums()


    def getVal(self,
               ncase : int,
               ) -> InVal:
        """
        Get the input value for a specific case.

        Parameters
        ----------
        ncase : int
            The number of the case to get the value for.

        Returns
        -------
        val : monaco.mc_val.InVal
            The input value for that case.
        """
        ismedian = False
        if (ncase == 0) and self.firstcaseismedian:
            ismedian = True

        val = InVal(name=self.name, ncase=ncase,
                    pct=self.pcts[ncase], num=self.nums[ncase].item(),
                    dist=self.dist, nummap=self.nummap,
                    ismedian=ismedian)
        return val


    def getDistMedian(self) -> float:
        """
        Get the median value for the statistical distribution.

        Returns
        -------
        median : float
            The median value of that distribution.
        """
        dist = self.dist(**self.distkwargs)
        median = dist.ppf(0.5)
        return median


    def getDistMean(self) -> float:
        """
        Get the mean value (also called the expected value) for the statistical
        distribution.

        Returns
        -------
        mean : float
            The mean value of that distribution.
        """
        dist = self.dist(**self.distkwargs)
        ev = dist.expect()

        if isinstance(self.dist, rv_continuous):
            return ev

        # For a discrete distribution, we take the nearest discrete value
        # closest to the expected value
        elif isinstance(self.dist, rv_discrete):
            eps = np.finfo(float).eps
            p = dist.cdf(ev)
            ev_candidates = dist.ppf([p - eps, p, p + eps])
            ev_candidates_dist = abs(ev_candidates - ev)
            ev_closest = ev_candidates[np.nanargmin(ev_candidates_dist)]
            return ev_closest

        else:
            return None



### OutVar Class ###
class OutVar(Var):
    """
    A Monte Carlo output variable.

    Parameters
    ----------
    name : str
        The name of this value.
    vals : list[Any]
        The values returned from the simulation.
    valmap : dict, default: None
        A dictionary mapping nonnumeric values to numbers.
    ndraws : int
        The number of random draws.
    seed : int, default: np.random.get_state(legacy=False)['state']['key'][0]
        The random seed for bootstrapping.
    firstcaseismedian : bool, default: False
        Whether the first case represents the median case.
    datasource : str, default: None
        If the outvals were imported from a file, this is the filepath. If
        generated through monaco, then None.

    Attributes
    ----------
    isscalar : bool
        Whether this is a scalar variable. Alway True for an input variable.
    maxdim : int
        The maximum dimension of the values. Always 0 for an input variable.
    nummap : dict
        A dictionary mapping numbers to nonnumeric values (the inverse of
        `valmap`).
    nums : list[np.ndarry]
        The numbers corresponding to the output values. If valmap is None, then
        `nums == list[np.array(vals)]`.
    varstats : list[moncao.VarStat.VarStat]
        A list of all the variable statistics for this variable.
    """
    def __init__(self,
                 name              : str,
                 vals              : list[Any],
                 valmap            : dict = None,
                 ndraws            : int  = None,
                 seed              : int = np.random.get_state(legacy=False)['state']['key'][0],
                 firstcaseismedian : bool = False,
                 datasource        : Optional[str] = None,
                 ):
        if ndraws is None:
            ndraws = len(vals)
            if firstcaseismedian:
                ndraws = ndraws - 1

        super().__init__(name=name, ndraws=ndraws, seed=seed,
                         firstcaseismedian=firstcaseismedian,
                         datasource=datasource)
        self.vals = vals
        self.valmap = valmap
        if valmap is None:
            self.extractValMap()
        self.genNumMap()
        self.mapVals()
        self.genMaxDim()
        self.sensitivity_indices : None | dict = None
        self.sensitivity_ratios  : None | dict = None


    def genMaxDim(self) -> None:
        """
        Parse the output values to determine the maximum dimension of each of
        their shapes.
        """
        self.maxdim = 0
        for num in self.nums:
            self.maxdim = max(self.maxdim, len(num.shape))

        self.isscalar = False
        if self.maxdim == 0:
            self.isscalar = True


    def extractValMap(self) -> None:
        """
        Parse the output values and extract a valmap.
        """
        Val0 = self.getVal(0)
        if Val0.valmap is not None:
            if Val0.valmapsource == 'auto':
                uniquevals : set[Any] = set()
                for i in range(self.ncases):
                    uniquevals.update(self.getVal(i).valmap.keys())
                self.valmap = dict()
                for i, val in enumerate(sorted(uniquevals)):
                    self.valmap[val] = i
            else:
                self.valmap = Val0.valmap


    def genNumMap(self) -> None:
        """
        Generate the nummap by inverting `valmap`.
        """
        if self.valmap is None:
            self.nummap = None
        else:
            self.nummap = {num: val for val, num in self.valmap.items()}


    def mapVals(self) -> None:
        """
        Generate `nums` by mapping the values with their valmap.
        """
        self.nums = []
        for i in range(self.ncases):
            self.nums.append(np.array(self.getVal(i).num))


    def getVal(self, ncase : int) -> OutVal:
        """
        Get the variable value for a specific case.

        Parameters
        ----------
        ncase : int
            The number of the case to get the value for.

        Returns
        -------
        val : monaco.mc_val.OutVal
            The output value.
        """
        ismedian = False
        if (ncase == 0) and self.firstcaseismedian:
            ismedian = True

        val = OutVal(name=self.name, ncase=ncase, val=self.vals[ncase],
                     valmap=self.valmap, ismedian=ismedian)
        return val


    def getMedianVal(self) -> OutVal:
        """
        Get the median value for this output variable if `firstcaseismedian`.

        Returns
        -------
        val : monaco.mc_val.OutVal
            The median output value.
        """
        val = None
        if self.firstcaseismedian:
            val = OutVal(name=self.name, ncase=0, val=self.vals[0],
                         valmap=self.valmap, ismedian=True)
        return val


    def split(self) -> dict[str, 'OutVar']:  # Quotes in typing to avoid import error
        """
        Split a multidimentional output variable along its outermost dimension,
        and generate individual OutVar objects for each index.

        Returns
        -------
        vars : dict[str : monaco.mc_var.OutVar]
        """
        vars : dict[str, 'OutVar'] = dict()
        if self.maxdim > 0:
            # First ensure that the vals have the same shape over all cases
            shape = self.nums[0].shape
            for num in self.nums:
                if num.shape[0] != shape[0]:
                    return vars

            for i in range(shape[0]):
                name = self.name + f' [{i}]'
                vals = []
                for j in range(self.ncases):
                    vals.append(self.vals[j][i])
                vars[name] = OutVar(name=name, vals=vals, ndraws=self.ndraws,
                                    valmap=self.valmap,
                                    firstcaseismedian=self.firstcaseismedian)
                for varstat in self.varstats:
                    vars[name].addVarStat(stat=varstat.stat,
                                          statkwargs=varstat.statkwargs,
                                          name=varstat.name)
        return vars


    def plotSensitivities(self,
                          sensitivities : Sensitivities = Sensitivities.RATIOS,
                          ax            : Optional[Axes] = None,
                          title         : str            = '',
                          ) -> tuple[Figure, Axes]:
        """
        Plot the sensitivity indices for this variable.
        See monaco.mc_plot.plot_sensitivities() for API details.
        """
        fig, ax = plot_sensitivities(outvar=self, sensitivities=sensitivities,
                                     ax=ax, title=title)
        return fig, ax



# gaussian_statistics.py
from __future__ import annotations

import scipy.stats
import numpy as np
from monaco.mc_enums import StatBound


def pct2sig(p     : float,
            bound : StatBound = StatBound.TWOSIDED,
            ) -> float:
    """
    Converts a percentile to a gaussian sigma value (1-sided), or to the
    sigma value for which the range (-sigma, +sigma) bounds that percent of
    the normal distribution (2-sided).

    Parameters
    ----------
    p : float
        The percentile to convert, 0 < p < 1.
    bound : monaco.mc_enums.StatBound
        The statistical bound, either '1-sided' or '2-sided'.

    Returns
    -------
    sig : float
        The gaussian sigma value for the input percentile.
    """
    sig = None
    if p <= 0 or p >= 1:
        raise ValueError(f'{p=} must be 0 < p < 1')

    if bound == StatBound.TWOSIDED:
        if p >= 0.5:
            sig = scipy.stats.norm.ppf(1-(1-p)/2)
        else:
            sig = -scipy.stats.norm.ppf(p/2)
    elif bound == StatBound.ONESIDED:
        sig = scipy.stats.norm.ppf(p)
    else:
        raise ValueError(f"{bound=} must be {StatBound.ONESIDED} or {StatBound.TWOSIDED}")

    return sig



def sig2pct(sig   : float,
            bound : StatBound = StatBound.TWOSIDED,
            ) -> float:
    """
    Converts a gaussian sigma value to a percentile (1-sided), or to the
    percent of the normal distribution bounded by (-sigma, +sigma) (2-sided).

    Parameters
    ----------
    sig : float
        The gaussian sigma value to convert.
    bound : monaco.mc_enums.StatBound
        The statistical bound, either '1-sided' or '2-sided'.

    Returns
    -------
    p : float
        The corresponding percentile or percent, 0 < p < 1.
    """
    p = None
    if bound == StatBound.TWOSIDED:
        p = 1-(1-scipy.stats.norm.cdf(sig))*2
    elif bound == StatBound.ONESIDED:
        p = scipy.stats.norm.cdf(sig)
    else:
        raise ValueError(f"{bound=} must be {StatBound.ONESIDED} or {StatBound.TWOSIDED}")

    return p



def conf_ellipsoid_pct2sig(p  : float,
                           df : int,
                           ) -> float:
    """
    Converts a percentile to a sigma value which bounds a df-dimensional
    gaussian distribution, used in generating confidence ellipsoids. Note
    that in the 1-D case, np.sqrt(scipy.stats.chi2.ppf(p, df=1)) is
    equivalent to pct2sig(p >= 0.5, bound = '2-sided') ==
    scipy.stats.norm.ppf(1-(1-p)/2)

    Parameters
    ----------
    p : float
        The percentile to convert, 0 < p < 1.
    df : int
        The degrees of freedom, df > 0.

    Returns
    -------
    sig : float
        The gaussian sigma value for the input percentile.
    """
    sig = None

    if p <= 0 or p >= 1:
        raise ValueError(f'{p=} must be 0 < p < 1')
    elif df <= 0:
        raise ValueError(f'{df=} must be > 0')
    else:
        sig = np.sqrt(scipy.stats.chi2.ppf(p, df=df))

    return sig



def conf_ellipsoid_sig2pct(sig : float,
                           df  : int,
                           ) -> float:
    """
    Converts a sigma value which bounds a df-dimensional gaussian distribution,
    to a percentil used in generating confidence ellipsoids. Note that in the
    1-D case, scipy.stats.chi2.cdf(sig**2, df=1) is equivalent to
    sig2pct(sig > 0, bound='2-sided) == 1-(1-scipy.stats.norm.cdf(sig))*2

    Parameters
    ----------
    sig : float
        The gaussian sigma value to convert, sig > 0.
    df : int
        The degrees of freedom, df > 0.

    Returns
    -------
    p : float
        The corresponding percentile, 0 < p < 1.
    """

    if sig <= 0:
        raise ValueError(f'{sig=} must be > 0')
    elif df <= 0:
        raise ValueError(f'{df=} must be > 0')

    p = scipy.stats.chi2.cdf(sig**2, df=df)
    return p



# mc_val.py
from __future__ import annotations

import numpy as np
from monaco.helper_functions import is_num, hashable_val, flatten
from typing import Any
from scipy.stats import rv_discrete, rv_continuous
from abc import ABC
try:
    import pandas as pd
except ImportError:
    pd = None


### Val Base Class ###
class Val(ABC):
    """
    Abstract base class to hold the data for a Monte Carlo value.

    Parameters
    ----------
    name : str
        The name of this value.
    ncase : int
        The number of the case for this value.
    ismedian : bool
        Whether this case represents the median case.
    """
    def __init__(self,
                 name     : str,
                 ncase    : int,
                 ismedian : bool,
                 ):
        self.name = name
        self.ncase = ncase
        self.ismedian = ismedian

        self.val      : Any
        self.valmap   : dict[Any, float]
        self.num      : np.float_ | np.ndarray
        self.nummap   : dict[float, Any]
        self.isscalar : bool
        self.shape    : tuple[int, ...]



### InVal Class ###
class InVal(Val):
    """
    A Monte Carlo input value.

    Parameters
    ----------
    name : str
        The name of this value.
    ncase : int
        The number of the case for this value.
    pct : float
        The percentile of the value draw.
    num : float
        The number corresponding to the statistical percentile draw.
    dist : scipy.stats.rv_discrete | scipy.stats.rv_continuous
        The statistical distribution that `num` was drawn from.
    nummap : dict[float, Any], default: None
        A dictionary mapping numbers to nonnumeric values.
    ismedian : bool, default: False
        Whether this case represents the median case,

    Attributes
    ----------
    val : Any
        The value corresponding to the drawn number. If `nummap` is None, then
        this is equal to `num`.
    isscalar : bool
        Whether the value is scalar.
    shape : tuple[int]
        The shape of the value.
    valmap : dict[Any, float]
        A dictionary mapping nonnumeric values to numbers (the inverse of
        `nummap`).
    """
    def __init__(self,
                 name     : str,
                 ncase    : int,
                 pct      : float,
                 num      : float,
                 dist     : rv_discrete | rv_continuous,
                 nummap   : dict[float, Any] | None = None,
                 ismedian : bool = False,
                 ):

        super().__init__(name=name, ncase=ncase, ismedian=ismedian)
        self.dist = dist
        self.pct = pct
        self.num = np.float_(num)
        self.nummap = nummap
        self.isscalar = True
        self.shape = ()

        self.mapNum()
        self.genValMap()


    def __repr__(self):
        return (f"{self.__class__.__name__}('{self.name}', ncase={self.ncase}, " +
                f"val={self.val} ({self.num}), pct={self.pct:0.4f})")


    def mapNum(self) -> None:
        """
        Generate `val` based on the drawn number and the nummap.
        """
        if self.nummap is None:
            self.val = self.num
        else:
            self.val = self.nummap[self.num.item()]


    def genValMap(self) -> None:
        """
        Generate the valmap based on the nummap.
        """
        if self.nummap is None:
            self.valmap = None
        else:
            self.valmap = {hashable_val(val): num for num, val in self.nummap.items()}



### OutVal Class ###
class OutVal(Val):
    """
    A Monte Carlo output value.

    Parameters
    ----------
    name : str
        The name of this value.
    ncase : int
        The number of the case for this value.
    val : float
        The output value.
    valmap : dict[Any, float], default: None
        A dictionary mapping nonnumeric values to numbers.
    ismedian : bool, default: False
        Whether this case represents the median case,

    Attributes
    ----------
    num : np.array
        A number corresponding to the output value. If `valmap` is None and val
        is nonnumeric, then will be an integer as assigned by extractValMap().
    valmapsource : str
        Either 'assigned' or 'auto' based on whether a valmap was passed in.
    isscalar : bool
        Whether the value is scalar.
    shape : tuple[int]
        The shape of the value.
    nummap : dict[float, Any]
        A dictionary mapping numbers to nonnumeric values (the inverse of
        `valmap`).
    """
    def __init__(self,
                 name     : str,
                 ncase    : int,
                 val      : Any,
                 valmap   : dict[Any, float] | None = None,
                 ismedian : bool = False,
                 ):
        super().__init__(name=name, ncase=ncase, ismedian=ismedian)
        self.val = val
        self.valmap = valmap
        self.convertPandas()

        self.genShape()
        if valmap is None:
            self.valmapsource = 'auto'
            self.extractValMap()
        else:
            self.valmapsource = 'assigned'
        self.mapVal()
        self.genNumMap()


    def __repr__(self):
        return (f"{self.__class__.__name__}('{self.name}', ncase={self.ncase}, " +
                f"val={self.val} ({self.num}))")


    def convertPandas(self) -> None:
        """
        If the output value is a pandas dataseries or index, convert it to a
        format we understand.
        """
        if pd:
            if isinstance(self.val, pd.Series) or isinstance(self.val, pd.Index):
                self.val = self.val.values


    def genShape(self) -> None:
        """
        Calculate the shape of the output value, and whether it is a scalar.
        """
        try:
            vals_array = np.array(self.val, dtype='float')
        except ValueError:
            vals_array = np.array(self.val, dtype='object')
        self.shape = vals_array.shape

        self.isscalar = False
        if self.shape == ():
            self.isscalar = True


    def extractValMap(self) -> None:
        """
        Parse the output value and extract a valmap.
        """
        vals_flattened = flatten([self.val])
        if all((isinstance(x, bool) or isinstance(x, np.bool_))
               for x in vals_flattened):
            self.valmap = {True: 1, False: 0}
        elif any(not is_num(x) for x in vals_flattened):
            sorted_vals = sorted(set(hashable_val(x) for x in vals_flattened))
            self.valmap = {val: idx for idx, val in enumerate(sorted_vals)}


    def mapVal(self) -> None:
        """
        Map the output value to a number or array of numbers.
        """
        if self.valmap is None:
            self.num = np.array(self.val)
        elif self.isscalar:
            self.num = np.array(self.valmap[hashable_val(self.val)])
        else:
            num = np.array(self.val, dtype='object')
            if len(self.shape) == 1:
                for i in range(self.shape[0]):
                    num[i] = self.valmap[hashable_val(self.val[i])]
            else:
                for i in range(self.shape[0]):
                    for j in range(self.shape[1]):
                        num[i][j] = self.valmap[hashable_val(self.val[i][j])]
            self.num = np.array(num, dtype='float')


    def genNumMap(self) -> None:
        """
        Invert the valmap to get a nummap.
        """
        if self.valmap is None:
            self.nummap = None
        else:
            self.nummap = {hashable_val(np.array(num)): val for val, num in self.valmap.items()}


    def split(self) -> dict[str, 'OutVal']:  # Quotes in typing to avoid import error
        """
        Split a multidimentional output value along its outermost dimension,
        and generate individual OutVal objects for each index.

        Returns
        -------
        vals : dict[str : monaco.mc_val.OutVal]
        """
        vals = dict()
        if len(self.shape) > 1:
            for i in range(self.shape[0]):
                name = self.name + f' [{i}]'
                vals[name] = OutVal(name=name, ncase=self.ncase, val=self.val[i],
                                    valmap=self.valmap, ismedian=self.ismedian)
        return vals



# dvars_sensitivity.py
from __future__ import annotations

# Somewhat hacky type checking to avoid circular imports:
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from monaco.mc_sim import Sim

import numpy as np
from monaco.helper_functions import vprint
from scipy.optimize import minimize
from scipy.linalg import det
from warnings import warn

# numba is recommended for speed, as this will be very slow otherwise
try:
    from numba import jit
except ImportError:
    warn("Consider installing numba for better performance", UserWarning)

    # create a dummy decorator
    def jit(f=None, *args, **kwargs):
        def decorator(func):
            return func

        if callable(f):
            return f
        else:
            return decorator


def calc_sensitivities(sim        : 'Sim',
                       outvarname : str,
                       Hj         : float = 1.0,
                       tol        : float = 1e-6,
                       verbose    : bool  = False,
                       ) -> tuple[np.ndarray, np.ndarray]:
    """
    Calculates the global sensitivity indices and ratios for a specific output
    variable to each of a simulation's input variables.

    This implements the D-VARS algorithm as dscribed in Reference [1]_ to
    calculate global sensitivity indices from a set of given data. The
    implementation largely follows the Reference's supplementary notes, though
    Multi-SQP is replaced with scipy's minimize function implementing L-BFGS-B.
    See also Reference [2]_ for some theoretical background on the VARS and
    IVARS methods of calculating sensitivity indices.

    Parameters
    ----------
    sim : monaco.mc_sim.Sim
        The input simulation.
    outvarname : str
        The name of the output variable to calculate sensitivities for. Note
        that the output variable must be scalar.
    Hj : float, default: 1.0
        The fraction of the total parameter space to integrate over. Note that
        the linear correlation function only has one hyperparameter, so as the
        Reference notes it is unable to distinguish variogram effects at varying
        length scales. So, this should not be set to anything other than 1 in
        practice.
    tol : float, default 1e-6
        The convergence tolerance for scipy's minimize function acting on the
        negative log likelihood function.
    verbose : bool, default False
        Whether to print diagnostic information.

    Returns
    -------
    sensitivities : numpy.ndarray
        The global sensitivity indices for the output variable for each of the
        sim's input variables.
    ratios : numpy.ndarray
        The global sensitivity ratios for the output variable for each of the
        sim's input variables, essentially the fraction of each input variable's
        ability to explain the output variance.

    References
    ----------
    .. [1] Sheikholeslami, Razi, and Saman Razavi. "A fresh look at variography:
           measuring dependence and possible sensitivities across geophysical
           systems from any given data." Geophysical Research Letters 47.20
           (2020): e2020GL089829.
    .. [2] Razavi, Saman, and Hoshin V. Gupta. "A new framework for
           comprehensive, robust, and efficient global sensitivity analysis:
           1. Theory." Water Resources Research 52.1 (2016): 423-439.
    """
    phi_opt = calc_phi_opt(sim, outvarname, tol, verbose)

    variance = np.var(np.array([sim.outvars[outvarname].nums]))
    sensitivities = np.zeros(sim.ninvars)
    for j in range(sim.ninvars):
        sensitivities[j] = calc_Gammaj(Hj, phi_opt[j], variance)
    ratios = sensitivities/sum(sensitivities)

    return sensitivities, ratios


def calc_phi_opt(sim        : 'Sim',
                 outvarname : str,
                 tol        : float = 1e-6,
                 verbose    : bool = False
                 ) -> np.ndarray:
    """
    Calculate the optimal hyperparameters for the covariance functions between
    the output variable and each of the input variables via maximum likelihood
    estimation (MLE). MLE works by minimizing a negative log-likelihood
    function.

    Parameters
    ----------
    sim : monaco.mc_sim.Sim
        The input simulation.
    outvarname : str
        The name of the output variable to calculate sensitivities for. Note
        that the output variable must be scalar.
    tol : float, default 1e-6
        The convergence tolerance for scipy's minimize function acting on the
        negative log likelihood function.
    verbose : bool, default False
        Whether to print diagnostic information.

    Returns
    -------
    phi_opt : numpy.ndarray
        The learned hyperparameters for the covariance functions.
    """
    phi_max = 1e6
    phi_min = 0
    phi0 = 1

    phi0s = []
    bounds = []
    for _ in range(sim.ninvars):
        phi0s.append(phi0)
        bounds.append((phi_min, phi_max))

    vprint(sim.verbose, 'Calculating optimal hyperparameters Φ for ' +
                       f"'{outvarname}' covariances...")
    X, Y = full_states(sim, outvarname)
    res = minimize(L_runner, phi0s, args=(X, Y, verbose), bounds=bounds,
                   tol=tol, method='L-BFGS-B')
    phi_opt = res.x
    vprint(sim.verbose, 'Done calculating optimal hyperparameters.')

    return phi_opt


def full_states(sim : 'Sim',
                outvarname : str,
                ) -> tuple[np.ndarray, np.ndarray]:
    """
    Get the full input and output states in combined matrices.

    Parameters
    ----------
    sim : monaco.mc_sim.Sim
        The input simulation.
    outvarname : str
        The name of the output variable to calculate sensitivities for. Note
        that the output variable must be scalar.

    Returns
    -------
    X : numpy.ndarray
        The input variables' percentiles.
    Y : numpy.ndarray
        The output variable's nums.
    """
    X = np.zeros((sim.ncases, sim.ninvars))
    Y = np.zeros((sim.ncases, 1))
    for i, varname in enumerate(sim.invars):
        X[:, i] = sim.invars[varname].pcts
    Y[:, 0] = sim.outvars[outvarname].nums
    return X, Y


def calc_Gammaj(Hj       : float,
                phij     : float,
                variance : float
                ) -> float:
    """
    Calculates the IVARS sensitivity index from a learned covariance function.

    This integrates the directional variogram for a specific input variable
    using trapezoidal integration.

    Parameters
    ----------
    Hj : float
        The fraction of the total parameter space to integrate over. Note that
        the linear correlation function only has one hyperparameter, so as the
        Reference notes it is unable to distinguish variogram effects at varying
        length scales. So, this should not be set to anything other than 1 in
        practice.
    phij : float
        The learned hyperparameter for the covariance function between the
        output and input variable.
    variance : float
        The variance of the output variable.

    Returns
    -------
    Gammaj : float
        The global sensitivity index for this output-input variable pair.
    """
    dh = 1e-3
    q = int(np.floor(Hj/dh))
    rjs = np.zeros(q+1)
    for i in range(q+1):
        rjs[i] = calc_rj(dh*i, phij)

    Gammaj = np.trapz(1 - rjs) * dh * variance
    return Gammaj


def L_runner(phi     : np.ndarray,
             X       : np.ndarray,
             Y       : np.ndarray,
             verbose : bool = False
             ) -> float:
    """
    A wrapper function for calculating the negative log-likelihood cost.

    Parameters
    ----------
    phi : numpy.ndarray
        The hyperparameters for the covariance function.
    X : numpy.ndarray
        The state matrix for all the input variables percentiles.
    Y : numpy.ndarray
        The state matrix for the output variables nums.
    verbose : bool
        Whether to print the values at each step.

    Returns
    -------
    L : float
        The negative log-likelihood cost.
    """
    L = calc_L(phi, X, Y)
    vprint(verbose, f'L = {L:0.4f}, Φ = {phi}')
    return L


@jit(nopython=True, cache=True)
def calc_L(phi : np.ndarray,
           X   : np.ndarray,
           Y   : np.ndarray
           ) -> float:
    """
    Calculate the negative log-likelihood cost. Note that this is just-in-time
    compiled by numba for increased speed.

    Parameters
    ----------
    phi : numpy.ndarray
        The hyperparameters for the covariance function.
    X : numpy.ndarray
        The state matrix for all the input variables percentiles.
    Y : numpy.ndarray
        The state matrix for the output variables nums.

    Returns
    -------
    L : float
        The negative log-likelihood cost.
    """
    m = len(Y)
    M = np.ones((m, 1))
    R = calc_R(phi, X)
    Rinv = np.linalg.inv(R)
    Rdet = max(det(R), 1e-12)  # Protect for poor conditioning

    mu = np.linalg.inv(M.T @ Rinv @ M) @ (M.T @ Rinv @ Y)

    L_inner = Y - M*mu
    L = np.log(Rdet)/m + m*np.log(L_inner.T @ Rinv @ L_inner)
    L = L[0][0]
    return L


@jit(nopython=True, cache=True)
def calc_R(phi : np.ndarray,
           X   : np.ndarray,
           ) -> np.ndarray:
    """
    Calculate the correlation matrix between each of the input states.
    Note that this is just-in-time compiled by numba for increased speed.

    Parameters
    ----------
    phi : numpy.ndarray
        The hyperparameters for the covariance function.
    X : numpy.ndarray
        The state matrix for all the input variables percentiles.

    Returns
    -------
    R : float
        The correlation matrix.
    """
    m = X.shape[0]
    R = np.ones((m, m))
    for u in range(1, m):
        # do lower triangle only and duplicate across diag
        # diag will be all 1s
        for w in range(1, u):
            Ruw = calc_Ruw(phi, X[u, :], X[w, :])
            R[u, w] = Ruw
            R[w, u] = Ruw
    return R


@jit(nopython=True, cache=True)
def calc_Ruw(phi : np.ndarray,
             Xu  : np.ndarray,
             Xw  : np.ndarray
             ) -> float:
    """
    Calculate the correlation between two input states.
    Note that this is just-in-time compiled by numba for increased speed.

    Parameters
    ----------
    phi : numpy.ndarray
        The hyperparameters for the covariance function.
    Xu : numpy.ndarray
        The u'th input state.
    Xw : numpy.ndarray
        The w'th input state.

    Returns
    -------
    Ruw : float
        The correlation between the two states.
    """
    h = Xu - Xw
    Ruw = 1
    for j, hj in enumerate(h):
        Ruw = Ruw*calc_rj(hj, phi[j])
    return Ruw


@jit(nopython=True, cache=True)
def calc_rj(hj   : float,
            phij : float
            ) -> float:
    """
    The covariance function (also called a kernel). We currently use a linear
    kernel which has a single hyperparameter that must be learned.
    Note that this is just-in-time compiled by numba for increased speed.

    Parameters
    ----------
    hj : float
        The distance between two state elements.
    phij : float
        The hyperparameter for the function.

    Returns
    -------
    rj : float
        The covariance.
    """
    rj = max(0, 1 - phij*abs(hj))  # linear covariance function
    # rj = np.exp(-(abs(hj)/phij))  # exponential covariance function
    # rj = np.exp(-(hj/phij)**2)  # squared exponential covariance function
    return rj



# mc_varstat.py
from __future__ import annotations

# Somewhat hacky type checking to avoid circular imports:
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from monaco.mc_var import Var

import numpy as np
from copy import copy
from statistics import mode
from scipy.stats import bootstrap, moment, skew, kurtosis
from scipy.stats.mstats import gmean
from monaco.helper_functions import get_list
from monaco.gaussian_statistics import pct2sig, sig2pct
from monaco.order_statistics import (order_stat_TI_n, order_stat_TI_k,
                                     order_stat_P_k, get_iP)
from monaco.mc_enums import StatBound, VarStatType, VarStatSide
from typing import Any, Callable


class VarStat:
    """
    A variable statistic for a Monte Carlo variable.

    Parameters
    ----------
    var : monaco.mc_var.Var
        The variable to generate statistics for.
    stat : monaco.mc_enums.VarStatType | Callable
        The statistic to generate. Can be custom. If custom, must be able to
        accept an "axis" kwarg for bootstrap vectorization.
    statkwargs : dict[str:Any]
        The keyword arguments for the variable statistic.
    bootstrap : bool (default: True)
        Whether to use bootstrapping to generate confidence intervals for the
        statistic.
    bootstrap_k : int (default: 10)
        The k'th order statistic to determine the number of bootstrap draws for
        the given confidence level. Must be >= 1. Set higher for a smoother
        bootstrap distribution.
    conf : float (default: 0.95)
        The confidence level for the confidence interval.
    seed : int (default: np.random.get_state(legacy=False)['state']['key'][0])
        The random seed to use for bootstrapping.
    name : str
        The name of the variable statistic.

    Attributes
    ----------
    nums : numpy.ndarray
        The output of the variable statistic function applied to `var.nums`
    confidence_interval_low_nums : numpy.ndarray
        The nums for the low side of the confidence interval (None if no CI).
    confidence_interval_high_nums : numpy.ndarray
        The nums for the high side of the confidence interval (None if no CI).
    vals : list[Any]
        The values for the `nums` as determined by `var.nummap`
    confidence_interval_low_vals : list[Any]
        The values for `confidence_interval_low_nums` via `var.nummap`
    confidence_interval_high_vals : list[Any]
        The values for `confidence_interval_high_nums` via `var.nummap`
    bootstrap_n : int
        The number of bootstrap samples.

    Notes
    -----
    These are the valid stats with their statkwargs

    max()
        'max', no kwargs
    min()
        'min', no kwargs
    median()
        'median', no kwargs
    mean()
        'mean', no kwargs
    geomean()
        'geomean', no kwargs
    mode()
        'mode', no kwargs
    variance()
        'variance', no kwargs
    skewness()
        'skewness', no kwargs
    kurtosis()
        'kurtosis', no kwargs
    moment(n : int)
        'moment'
        `n` is the n'th moment, `n > 0`.
    percentile(p : float)
        'percentile'
        `p` is the percentile, `0 < p < 1`.
    sigma(sig : float, bound : monaco.mc_enums.StatBound)
        'sigma'
        `sig` is the gaussian sigma value, `-inf < sig < inf`.

        `bound` is the statistical bound, ether `'1-sided'` or `'2-sided'`.
        Default is `'2-sided'`.
    gaussianP(p : float, bound : monaco.mc_enums.StatBound)
        'gaussianP'
        `p` is the percentile, `0 < p < 1`.

        `bound` is the statistical bound, ether `'1-sided'` or `'2-sided'`.
        Default is `'2-sided'`.
    orderstatTI(p : float, c : float, bound : monaco.mc_enums.StatBound)
        'orderstatTI'
        `p` is the percentage, `0 < p < 1`

        `c` is the confidence, `0 < c < 1`. Default is `0.95`.

        `bound` is the statistical bound, ether `'1-sided'`, `'2-sided'`, or
        `'all'`. Default is `'2-sided'`.
    orderstatP(p : float, c : float, bound : monaco.mc_enums.StatBound)
        'orderstatP'
        `p` is the percentage, `0 < p < 1`

        `c` is the confidence, `0 < c < 1`. Default is `0.95`.

        `bound` is the statistical bound, ether `'1-sided lower'`,
        `'1-sided upper'`, `'2-sided'`, `'all'`, or '`nearest'`. Default is
        `'2-sided'`.
    """
    def __init__(self,
                 var         : Var,
                 stat        : str | VarStatType | Callable,
                 statkwargs  : dict[str, Any] | None = None,
                 bootstrap   : bool = True,
                 bootstrap_k : int = 10,
                 conf        : float = 0.95,
                 seed        : int = np.random.get_state(legacy=False)['state']['key'][0],
                 name        : str | None = None,
                 ):

        self.var = var
        if isinstance(stat, str):
            stat = stat.lower()
        self.stat = stat
        if statkwargs is None:
            statkwargs = dict()
        self.statkwargs = statkwargs

        self.nums : np.ndarray = np.array([])
        self.vals : list[Any] | np.ndarray = []
        self.name = name

        self.bootstrap = bootstrap
        if bootstrap_k < 1:
            raise ValueError(f'bootstrap_k = {bootstrap_k} must be >= 1')
        self.bootstrap_k = bootstrap_k
        self.conf = conf
        self.confidence_interval_low_nums : list | np.ndarray = None
        self.confidence_interval_high_nums : list | np.ndarray = None
        self.confidence_interval_low_vals : list[Any] | np.ndarray = []
        self.confidence_interval_high_vals : list[Any] | np.ndarray = []
        self.bootstrap_n : int | None = None
        self.seed = seed

        if isinstance(stat, Callable):
            self.setName(f'{self.var.name} {str(stat)}')
            self.genStatsFunction(fcn=stat, fcnkwargs=statkwargs)
        elif stat == VarStatType.MAX:
            self.setName(f'{self.var.name} Max')
            self.genStatsFunction(fcn=np.max)
        elif stat == VarStatType.MIN:
            self.setName(f'{self.var.name} Min')
            self.genStatsFunction(fcn=np.min)
        elif stat == VarStatType.MEDIAN:
            self.setName(f'{self.var.name} Median')
            self.genStatsFunction(fcn=np.median)
        elif stat == VarStatType.MEAN:
            self.setName(f'{self.var.name} Mean')
            self.genStatsFunction(fcn=np.mean)
        elif stat == VarStatType.GEOMEAN:
            self.setName(f'{self.var.name} Geometric Mean')
            self.genStatsFunction(fcn=gmean)
        elif stat == VarStatType.MODE:
            self.setName(f'{self.var.name} Mode')
            self.genStatsFunction(fcn=mode)
        elif stat == VarStatType.VARIANCE:
            self.setName(f'{self.var.name} Variance')
            self.genStatsFunction(fcn=np.var)
        elif stat == VarStatType.SKEWNESS:
            self.setName(f'{self.var.name} Skewness')
            self.genStatsFunction(fcn=skew)
        elif stat == VarStatType.KURTOSIS:
            self.setName(f'{self.var.name} Kurtosis')
            self.genStatsFunction(fcn=kurtosis)
        elif stat == VarStatType.MOMENT:
            self.genStatsMoment()
        elif stat == VarStatType.PERCENTILE:
            self.genStatsPercentile()
        elif stat == VarStatType.SIGMA:
            self.genStatsSigma()
        elif stat == VarStatType.GAUSSIANP:
            self.genStatsGaussianP()
        elif stat == VarStatType.ORDERSTATTI:
            self.genStatsOrderStatTI()
        elif stat == VarStatType.ORDERSTATP:
            self.genStatsOrderStatP()
        else:
            raise ValueError(f'{stat=} must be callable, or one of the following: ' +
                             f'{VarStatType.MAX}, {VarStatType.MIN}, {VarStatType.MEDIAN}, ' +
                             f'{VarStatType.MEAN}, {VarStatType.GEOMEAN}, {VarStatType.MODE}, ' +
                             f'{VarStatType.PERCENTILE}, {VarStatType.SIGMA}, ' +
                             f'{VarStatType.GAUSSIANP}, {VarStatType.ORDERSTATTI}, ' +
                             f'{VarStatType.ORDERSTATP}')


    def genStatsMoment(self) -> None:
        """
        Get the n'th moment about the mean of the variable.
        """
        if 'n' not in self.statkwargs:
            raise ValueError(f'{self.stat} requires the kwarg ''n''')

        self.n = self.statkwargs['n']
        self.setName(f'{self.var.name} {self.n}''th Moment')
        self.genStatsFunction(moment, {'moment': self.n})


    def genStatsPercentile(self) -> None:
        """
        Get the value of the variable at the inputted percentile.
        """
        if 'p' not in self.statkwargs:
            raise ValueError(f'{self.stat} requires the kwarg ''p''')

        self.p = self.statkwargs['p']
        self.setName(f'{self.var.name} {self.p*100}% Percentile')
        self.genStatsFunction(np.quantile, {'q': self.p})


    def genStatsSigma(self) -> None:
        """
        Get the value of the variable at the inputted sigma value, assuming
        a gaussian distribution.
        """
        if 'sig' not in self.statkwargs:
            raise ValueError(f'{self.stat} requires the kwarg ''sig''')
        if 'bound' not in self.statkwargs:
            self.bound = StatBound.TWOSIDED
        else:
            self.bound = self.statkwargs['bound']

        self.sig = self.statkwargs['sig']
        self.p = sig2pct(self.sig, bound=self.bound)
        self.setName(f'{self.var.name} {self.sig} Sigma')
        self.genStatsFunction(self.sigma, {'sig': self.sig})


    def genStatsGaussianP(self) -> None:
        """
        Get the value of the variable at the inputted percentile value,
        assuming a gaussian distribution.
        """
        if 'p' not in self.statkwargs:
            raise ValueError(f'{self.stat} requires the kwarg ''p''')
        if 'bound' not in self.statkwargs:
            self.bound = StatBound.TWOSIDED
        else:
            self.bound = self.statkwargs['bound']

        self.p = self.statkwargs['p']
        self.sig = pct2sig(self.p, bound=self.bound)
        self.setName(f'{self.var.name} Guassian {self.p*100}%')
        self.genStatsFunction(self.sigma, {'sig': self.sig})


    def sigma(self,
              x,  # TODO: explicit typing here
              sig  : float,
              axis : int | None = None,
              ) -> float:
        """
        Calculate the sigma value of a normally distributed list of numbers.

        Parameters
        ----------
        x : TODO typing
            The numbers to calculate the sigma value for.
        sig : float
            The sigma value.
        axis : int (default: None)
            The axis of x to calculate along.
        """
        std = np.std(x, axis=axis)
        return np.mean(x, axis=axis) + sig*std


    def statsFunctionWrapper(self,
                             x : Any,
                             axis : int | None = None,  # Needed for bootstrap vectorization
                             ) -> Any:
        """
        A wrapper function to allow using a bootstrap function that uses kwargs.
        Relies on self.fcn and self.fcnkwargs already being set. Note that fcn
        must accept an `axis` kwarg if bootstrapping.

        Parameters
        ----------
        x : Any
            The input for the function.
        axis : int (default: None)
            The axis of x to calculate along.
        """
        return self.fcn(x, **self.fcnkwargs, axis=axis)


    def genStatsFunction(self,
                         fcn       : Callable,
                         fcnkwargs : dict[str, Any] = None,
                         ) -> None:
        """
        A wrapper function to generate statistics via a generic function.

        Parameters
        ----------
        fcn : Callable
            The function used to generate the desired statistics.
        fcnkwargs : dict[str, Any]
            The keyword arguments for the function.
        """
        self.fcn = fcn
        if fcnkwargs is None:
            fcnkwargs = dict()
        self.fcnkwargs = fcnkwargs
        if self.bootstrap:
            self.bootstrap_n = order_stat_TI_n(self.bootstrap_k, p=0.5, c=self.conf)

        # Scalar Variables
        if self.var.isscalar:
            # Calculate nums and confidence interval for each point in the sequence
            self.nums = self.statsFunctionWrapper(self.var.nums)
            if self.bootstrap:
                # Switch to method='Bca' once https://github.com/scipy/scipy/issues/15883 resolved
                res = bootstrap((np.array(self.var.nums),), self.statsFunctionWrapper,
                                confidence_level=self.conf,
                                n_resamples=self.bootstrap_n,
                                random_state=self.seed, method='basic')
                self.confidence_interval_low_nums = res.confidence_interval.low
                self.confidence_interval_high_nums = res.confidence_interval.high

            # Calculate the corresponding vals based on the nummap
            self.vals = copy(self.nums)
            if self.bootstrap:
                self.confidence_interval_low_vals = copy(self.confidence_interval_low_nums)
                self.confidence_interval_high_vals = copy(self.confidence_interval_high_nums)
            if self.var.nummap is not None:
                self.vals = [self.var.nummap[num] for num in self.nums]
                if self.bootstrap:
                    self.confidence_interval_low_vals = \
                        [self.var.nummap[num] for num in self.confidence_interval_low_nums]
                    self.confidence_interval_high_vals = \
                        [self.var.nummap[num] for num in self.confidence_interval_high_nums]

        # 1-D Variables
        elif self.var.maxdim == 1:
            nums_list = get_list(self.var.nums)
            npoints = max(len(x) for x in nums_list)
            if self.bootstrap:
                confidence_interval_low_nums = []
                confidence_interval_high_nums = []

            # Calculate nums and confidence interval for each point in the sequence
            nums = []
            for i in range(npoints):
                numsatidx = np.array([x[i] for x in nums_list if len(x) > i])
                nums.append(self.statsFunctionWrapper(numsatidx))
                if self.bootstrap:
                    # Switch to Bca once https://github.com/scipy/scipy/issues/15883 resolved
                    res = bootstrap((numsatidx,), self.statsFunctionWrapper,
                                    confidence_level=self.conf,
                                    n_resamples=self.bootstrap_n,
                                    random_state=self.seed, method='basic')
                    confidence_interval_low_nums.append(res.confidence_interval.low)
                    confidence_interval_high_nums.append(res.confidence_interval.high)
            self.nums = nums
            if self.bootstrap:
                self.confidence_interval_low_nums = confidence_interval_low_nums
                self.confidence_interval_high_nums = confidence_interval_high_nums

            # Calculate the corresponding vals based on the nummap
            self.vals = copy(self.nums)
            if self.bootstrap:
                self.confidence_interval_low_vals = copy(self.confidence_interval_low_nums)
                self.confidence_interval_high_vals = copy(self.confidence_interval_high_nums)
            if self.var.nummap is not None:
                self.vals = [[self.var.nummap[x] for x in y] for y in self.nums]
                if self.bootstrap:
                    self.confidence_interval_low_vals \
                        = [[self.var.nummap[x] for x in y]
                           for y in self.confidence_interval_low_nums]
                    self.confidence_interval_low_vals \
                        = [[self.var.nummap[x] for x in y]
                           for y in self.confidence_interval_high_nums]

        else:
            # Suppress warning since this will become valid when Var is split
            # warn('VarStat only available for scalar or 1-D data')
            pass


    def genStatsOrderStatTI(self) -> None:
        """Get the order statistic tolerance interval value of the variable."""
        self.checkOrderStatsKWArgs()

        if self.bound == StatBound.ONESIDED and self.p >= 0.5:
            self.side = VarStatSide.HIGH
        elif self.bound == StatBound.ONESIDED:
            self.side = VarStatSide.LOW
        elif self.bound == StatBound.TWOSIDED:
            self.side = VarStatSide.BOTH
        elif self.bound == StatBound.ALL:
            self.bound = StatBound.TWOSIDED
            self.side = VarStatSide.ALL
        else:
            raise ValueError(f'{self.bound} is not a valid bound for genStatsOrderStatTI')

        if isinstance(self.bound, StatBound):
            bound_str = self.bound.value
        else:
            bound_str = self.bound

        self.setName(f'{self.var.name} ' +
                     f'{bound_str} P{round(self.p*100,4)}/{round(self.c*100,4)}% ' +
                      'Confidence Interval')

        self.k = order_stat_TI_k(n=self.var.ncases, p=self.p, c=self.c, bound=self.bound)

        if self.var.isscalar:
            sortednums = sorted(self.var.nums)
            if self.side == VarStatSide.LOW:
                sortednums.reverse()
            if self.side in (VarStatSide.HIGH, VarStatSide.LOW):
                self.nums = np.array(sortednums[-self.k])
                if self.var.nummap is not None:
                    self.vals = self.var.nummap[self.nums.item()]
            elif self.side == VarStatSide.BOTH:
                self.nums = np.array([sortednums[self.k-1], sortednums[-self.k]])
                if self.var.nummap is not None:
                    self.vals = np.array([self.var.nummap[self.nums[0]],
                                          self.var.nummap[self.nums[1]]])
            elif self.side == VarStatSide.ALL:
                self.nums = np.array([sortednums[self.k-1],
                                      np.median(sortednums),
                                      sortednums[-self.k]])
                if self.var.nummap is not None:
                    self.vals = np.array([self.var.nummap[self.nums[0]],
                                          self.var.nummap[self.nums[1]],
                                          self.var.nummap[self.nums[2]]])
            if self.var.nummap is None:
                self.vals = copy(self.nums)

        elif self.var.maxdim == 1:
            npoints = max(x.shape[0] if len(x.shape) > 0 else 0 for x in self.var.nums)
            self.nums = np.empty(npoints)
            if self.side == VarStatSide.BOTH:
                self.nums = np.empty((npoints, 2))
            elif self.side == VarStatSide.ALL:
                self.nums = np.empty((npoints, 3))
            for i in range(npoints):
                numsatidx = [x[i] for x in self.var.nums
                             if (len(x.shape) > 0 and x.shape[0] > i)]
                sortednums = sorted(numsatidx)
                if self.side == VarStatSide.LOW:
                    sortednums.reverse()
                if self.side in (VarStatSide.HIGH, VarStatSide.LOW):
                    self.nums[i] = sortednums[-self.k]
                elif self.side == VarStatSide.BOTH:
                    self.nums[i, :] = [sortednums[self.k - 1], sortednums[-self.k]]
                elif self.side == VarStatSide.ALL:
                    self.nums[i, :] = [sortednums[self.k - 1],
                                       sortednums[int(np.round(len(sortednums)/2)-1)],
                                       sortednums[-self.k]]
            if self.var.nummap is not None:
                self.vals = np.array([[self.var.nummap[x] for x in y] for y in self.nums])
            else:
                self.vals = copy(self.nums)

        else:
            # Suppress warning since this will become valid when Var is split
            # warn('VarStat only available for scalar or 1-D data')
            pass


    def genStatsOrderStatP(self) -> None:
        """Get the order statistic percentile value of the variable."""
        self.checkOrderStatsKWArgs()

        bound = self.bound
        if self.bound not in (StatBound.ONESIDED_UPPER, StatBound.ONESIDED_LOWER,
                              StatBound.TWOSIDED, StatBound.NEAREST, StatBound.ALL):
            raise ValueError(f'{self.bound} is not a valid bound for genStatsOrderStatP')
        elif self.bound in (StatBound.NEAREST, StatBound.ALL):
            bound = StatBound.TWOSIDED

        self.setName(f'{self.var.name} ' +
                     f'{self.bound} {self.c*100}% Confidence Bound around ' +
                     f'{self.p*100}th Percentile')

        self.k = order_stat_P_k(n=self.var.ncases, P=self.p, c=self.c, bound=bound)

        (iPl, iP, iPu) = get_iP(n=self.var.ncases, P=self.p)
        if self.var.isscalar:
            sortednums = sorted(self.var.nums)
            if self.bound == StatBound.ONESIDED_LOWER:
                self.nums = np.array(sortednums[iPl - self.k])
            elif self.bound == StatBound.ONESIDED_UPPER:
                self.nums = np.array(sortednums[iPu + self.k])
            elif self.bound == StatBound.NEAREST:
                self.nums = np.array(sortednums[iP])
            if self.bound in (StatBound.ONESIDED_LOWER,
                              StatBound.ONESIDED_UPPER,
                              StatBound.NEAREST):
                if self.var.nummap is not None:
                    self.vals = self.var.nummap[self.nums.item()]
            elif self.bound == StatBound.TWOSIDED:
                self.nums = np.array([sortednums[iPl - self.k], sortednums[iPu + self.k]])
                if self.var.nummap is not None:
                    self.vals = np.array([self.var.nummap[self.nums[0]],
                                          self.var.nummap[self.nums[1]]])
            elif self.bound == StatBound.ALL:
                self.nums = np.array([sortednums[iPl - self.k],
                                      sortednums[iP],
                                      sortednums[iPu + self.k]])
                if self.var.nummap is not None:
                    self.vals = np.array([self.var.nummap[self.nums[0]],
                                          self.var.nummap[self.nums[1]],
                                          self.var.nummap[self.nums[2]]])
            if self.var.nummap is None:
                self.vals = copy(self.nums)

        elif self.var.maxdim == 1:
            npoints = max(len(get_list(x)) for x in self.var.nums)
            self.nums = np.empty(npoints)
            if self.bound == StatBound.TWOSIDED:
                self.nums = np.empty((npoints, 2))
            elif self.bound == StatBound.ALL:
                self.nums = np.empty((npoints, 3))
            for i in range(npoints):
                numsatidx = [get_list(x)[i] for x in self.var.nums if len(get_list(x)) > i]
                sortednums = sorted(numsatidx)
                if self.bound == StatBound.ONESIDED_LOWER:
                    self.nums[i] = sortednums[iPl - self.k]
                elif self.bound == StatBound.ONESIDED_UPPER:
                    self.nums[i] = sortednums[iPu + self.k]
                elif self.bound == StatBound.NEAREST:
                    self.nums[i] = sortednums[iP]
                elif self.bound == StatBound.TWOSIDED:
                    self.nums[i, :] = [sortednums[iPl - self.k], sortednums[iPu + self.k]]
                elif self.bound == StatBound.ALL:
                    self.nums[i, :] = [sortednums[iPl - self.k],
                                       sortednums[iP],
                                       sortednums[iPu + self.k]]
            if self.var.nummap is not None:
                self.vals = np.array([[self.var.nummap[x] for x in y] for y in self.nums])
            else:
                self.vals = copy(self.nums)

        else:
            # Suppress warning since this will become valid when Var is split
            # warn('VarStat only available for scalar or 1-D data')
            pass


    def checkOrderStatsKWArgs(self) -> None:
        """Check the order statistic keyword arguments."""
        if 'p' not in self.statkwargs:
            raise ValueError(f'{self.stat} requires the kwarg ''p''')
        else:
            self.p = self.statkwargs['p']
        if 'c' not in self.statkwargs:
            self.c = 0.95
        else:
            self.c = self.statkwargs['c']
        if 'bound' not in self.statkwargs:
            self.bound = StatBound.TWOSIDED
        else:
            self.bound = self.statkwargs['bound']


    def setName(self,
                name : str,
                ) -> None:
        """
        Set the name for this variable statistic.

        Parameters
        ----------
        name : str
            The new name.
        """
        if self.name is None:
            self.name = name



# case_runners.py

from typing import Callable
from copy import copy
from monaco.mc_case import Case
from monaco.helper_functions import vwrite, vwarn, get_list
from datetime import datetime


def preprocess_case(preprocfcn: Callable,
                    case : Case,
                    debug : bool,
                    verbose : bool,
                    ) -> Case:
    """
    Preprocess a single Monte Carlo case.

    Parameters
    ----------
    case : monaco.mc_case.Case
        The case to preprocess.

    Returns
    -------
    case : monaco.mc_case.Case
        The same case, preprocessed.
    """
    case = copy(case)
    try:
        case.siminput = preprocfcn(case)
        case.haspreprocessed = True

    except Exception:
        if debug:
            raise
        else:
            vwarn(verbose, f'\nPreprocessing case {case.ncase} failed')

    return case


def run_case(runfcn: Callable,
             case : Case,
             debug : bool,
             verbose : bool,
             runsimid : int,
             ) -> Case:
    """
    Run a single Monte Carlo case.

    Parameters
    ----------
    case : monaco.mc_case.Case
        The case to run.

    Returns
    -------
    case : monaco.mc_case.Case
        The same case, run.
    """
    case = copy(case)

    try:
        case.starttime = datetime.now()
        case.simrawoutput = runfcn(*get_list(case.siminput))
        case.endtime = datetime.now()
        case.runtime = case.endtime - case.starttime
        case.runsimid = runsimid
        case.hasrun = True
        if not case.keepsiminput:
            case.siminput = ()

    except Exception:
        if debug:
            raise
        vwrite(verbose, f'\nRunning case {case.ncase} failed')

    return case


def postprocess_case(postprocfcn: Callable,
                     case : Case,
                     debug : bool,
                     verbose : bool,
                     ) -> Case:
    """
    Postprocess a single Monte Carlo case.

    Parameters
    ----------
    case : monaco.mc_case.Case
        The case to postprocess.

    Returns
    -------
    case : monaco.mc_case.Case
        The same case, postprocessed.
    """
    case = copy(case)
    try:
        postprocfcn(case, *get_list(case.simrawoutput))
        case.haspostprocessed = True
        if not case.keepsimrawoutput:
            case.simrawoutput = ()

    except Exception:
        if debug:
            raise
        else:
            vwrite(verbose, f'\nPostprocessing case {case.ncase} failed')

    return case



# order_statistics.py
from __future__ import annotations

import scipy.stats
import numpy as np
from monaco.mc_enums import StatBound


def order_stat_TI_n(k     : int,
                    p     : float,
                    c     : float,
                    nmax  : int       = int(1e7),
                    bound : StatBound = StatBound.TWOSIDED,
                    ) -> int:
    """
    For an Order Statistic Tolerance Interval, find the minimum n from k, p,
    and c

    Notes
    -----
    This function returns the number of cases n necessary to say that the true
    result of a measurement x will be bounded by the k'th order statistic with
    a probability p and confidence c. Variables l and u below indicate lower
    and upper indices of the order statistic.

    For example, if I want to use my 2nd highest measurement as a bound on 99%
    of all future samples with 90% confidence:

    .. code-block::

        n = order_stat_TI_n(k=2, p=0.99, c=0.90, bound='1-sided') = 389

    The 388th value of x when sorted from low to high, or sorted(x)[-2], will
    bound the upper end of the measurement with P99/90.

    '2-sided' gives the result for the measurement lying between the k'th lowest
    and k'th highest measurements. If we run the above function with
    bound='2-sided', then n = 668, and we can say that the true measurement lies
    between sorted(x)[1] and sorted(x)[-2] with P99/90.

    See chapter 5 of Reference [1]_ for statistical background.

    Parameters
    ----------
    k : int
        The k'th order statistic.
    p : float (0 < p < 1)
        The percent covered by the tolerance interval.
    c : float (0 < c < 1)
        The confidence of the interval bound.
    nmax : int, default: 1e7
        The maximum number of draws. Hard limit of 2**1000.
    bound : monaco.mc_enums.StatBound, default: '2-sided'
        The statistical bound, either '1-sided' or '2-sided'.

    Returns
    -------
    n : int
        The number of samples necessary to meet the constraints.

    References
    ----------
    .. [1] Hahn, Gerald J., and Meeker, William Q. "Statistical Intervals: A
       Guide for Practitioners." Germany, Wiley, 1991.
    """
    order_stat_var_check(p=p, k=k, c=c, nmax=nmax)

    if bound == StatBound.TWOSIDED:
        l = k  # we won't be using assymmetrical order stats
    elif bound == StatBound.ONESIDED:
        l = 0
    else:
        raise ValueError(f"{bound=} must be {StatBound.ONESIDED} or {StatBound.TWOSIDED}")

    # use bisection to get minimum n (secant method is unstable due to flat portions of curve)
    n = [1, nmax]
    maxsteps = 100  # nmax hard limit of 2^100
    u = n[1] + 1 - k
    if EPTI(n[1], l, u, p) < c:
        raise ValueError(f'n exceeded {nmax=} for P{100*p}/{c*100}. ' +
                          'Increase nmax or loosen constraints.')

    for i in range(maxsteps):
        step = (n[1]-n[0])/2
        ntemp = n[0] + np.ceil(step)
        if step < 1:
            return int(n[1])
        else:
            u = ntemp + 1 - k
            if EPTI(ntemp, l, u, p) <= c:
                n[0] = ntemp
            else:
                n[1] = ntemp
    raise ValueError(f'With {n=}, could not converge in {maxsteps=} steps. ' +
                    f'Is {nmax=} > 2^{maxsteps}?')



def order_stat_TI_p(n     : int,
                    k     : int,
                    c     : float,
                    ptol  : float     = 1e-9,
                    bound : StatBound = StatBound.TWOSIDED,
                    ) -> float:
    """
    For an Order Statistic Tolerance Interval, find the maximum p from n, k,
    and c.

    Parameters
    ----------
    n : int
        The number of samples.
    k : int
        The k'th order statistic.
    c : float (0 < c < 1)
        The confidence of the interval bound.
    ptol : float, default: 1e-9
        The absolute tolerance on determining p.
    bound : monaco.mc_enums.StatBound, default: '2-sided'
        The statistical bound, either '1-sided' or '2-sided'.

    Returns
    -------
    p : float (0 < p < 1)
        The percent which the tolerance interval covers corresponding to the
        input constraints.
    """
    order_stat_var_check(n=n, k=k, c=c)

    if bound == StatBound.TWOSIDED:
        l = k  # we won't be using assymmetrical order stats
    elif bound == StatBound.ONESIDED:
        l = 0
    else:
        raise ValueError(f"{bound=} must be {StatBound.ONESIDED} or {StatBound.TWOSIDED}")
    u = n + 1 - k

    # use bisection to get n (secant method is unstable due to flat portions of curve)
    p = [0.0, 1.0]
    maxsteps = 100  # p hard tolerance of 2^-100
    for i in range(maxsteps):
        step = (p[1]-p[0])/2
        ptemp = p[0] + step
        if step <= ptol:
            return p[1]
        else:
            if EPTI(n, l, u, ptemp) >= c:
                p[0] = ptemp
            else:
                p[1] = ptemp
    raise ValueError(f'With {p=}, could not converge under {ptol=} in {maxsteps} steps.')



def order_stat_TI_k(n     : int,
                    p     : float,
                    c     : float,
                    bound : StatBound = StatBound.TWOSIDED,
                    ) -> int:
    """
    For an Order Statistic Tolerance Interval, find the maximum k from n, p,
    and c.

    Parameters
    ----------
    n : int
        The number of samples.
    p : float (0 < p < 1)
        The percent covered by the tolerance interval.
    c : float (0 < c < 1)
        The confidence of the interval bound.
    bound : monaco.mc_enums.StatBound, default: '2-sided'
        The statistical bound, either '1-sided' or '2-sided'.

    Returns
    -------
    k : int
        The k'th order statistic.
    """
    order_stat_var_check(n=n, p=p, c=c)

    if bound == StatBound.TWOSIDED:
        l = 1  # we won't be using assymmetrical order stats
    elif bound == StatBound.ONESIDED:
        l = 0
    else:
        raise ValueError(f"{bound=} must be {StatBound.ONESIDED} or {StatBound.TWOSIDED}")

    if EPTI(n, l, n, p) < c:
        raise ValueError(f'{n=} is too small to meet {p=} at {c=} for {bound} ' +
                          'tolerance interval at any order statistic')

    # use bisection to get n (secant method is unstable due to flat portions of curve)
    k = [1, np.ceil(n/2)]
    maxsteps = 100  # nmax hard limit of 2^100
    for _ in range(maxsteps):
        step = (k[1]-k[0])/2
        ktemp = k[0] + np.ceil(step)
        if step < 1:
            return int(k[1])-1
        else:
            if bound == StatBound.TWOSIDED:
                l = ktemp  # we won't be using assymmetrical order stats
            elif bound == StatBound.ONESIDED:
                l = 0
            u = n + 1 - ktemp

            if EPTI(n, l, u, p) > c:
                k[0] = ktemp
            else:
                k[1] = ktemp
    raise ValueError(f'With {n=}, could not converge in {maxsteps} steps. Is n > 2^{maxsteps}?')



def order_stat_TI_c(n     : int,
                    k     : int,
                    p     : float,
                    bound : StatBound = StatBound.TWOSIDED,
                    ) -> float:
    """
    For an Order Statistic Tolerance Interval, find the maximum c from n, k,
    and p.

    Parameters
    ----------
    n : int
        The number of samples.
    k : int
        The k'th order statistic.
    p : float (0 < p < 1)
        The percent covered by the tolerance interval.
    bound : monaco.mc_enums.StatBound, default: '2-sided'
        The statistical bound, either '1-sided' or '2-sided'.

    Returns
    -------
    c : float (0 < c < 1)
        The confidence of the interval bound.
    """
    order_stat_var_check(n=n, p=p, k=k)

    if bound == StatBound.TWOSIDED:
        l = k  # we won't be using assymmetrical order stats
    elif bound == StatBound.ONESIDED:
        l = 0
    else:
        raise ValueError(f"{bound=} must be {StatBound.ONESIDED} or {StatBound.TWOSIDED}")

    u = n + 1 - k

    c = EPTI(n, l, u, p)
    return c



def order_stat_P_n(k     : int,
                   P     : float,
                   c     : float,
                   nmax  : int = int(1e7),
                   bound : StatBound = StatBound.TWOSIDED,
                   ) -> int:
    """
    Order Statistic Percentile, find minimum n from k, P, and c.

    Notes
    -----
    This function returns the number of cases n necessary to say that the true
    Pth percentile located at or between indices iPl and iPu of a measurement x
    will be bounded by the k'th order statistic with confidence c.

    For example, if I want to use my 5th nearest measurement as a bound on the
    50th Percentile with 90% confidence:

    .. code-block::

        n = order_stat_P_n(k=5, P=0.50, c=0.90, bound='2-sided') = 38
        iPl = np.floor(P*(n + 1)) = 19
        iPu = np.ceil(P*(n + 1)) = 20

    The 19-5 = 14th and 20+5= 25th values of x when sorted from low to high, or
    [sorted(x)[13], sorted(x)[24]] will bound the 50th percentile with 90%
    confidence.

    '2-sided' gives the upper and lower bounds. '1-sided lower' and
    '1-sided upper' give the respective lower or upper bound of the Pth
    percentile over the entire rest of the distribution.

    See chapter 5 of Reference [2]_ for statistical background.

    Parameters
    ----------
    k : int
        The k'th order statistic.
    P : float (0 < P < 1)
        The target percentile.
    c : float (0 < c < 1)
        The confidence of the interval bound.
    nmax : int, default: 1e7
        The maximum number of draws. Hard limit of 2**1000.
    bound : monaco.mc_enums.StatBound, default: '2-sided'
        The statistical bound, '1-sided upper', '1-sided lower', or '2-sided'.

    Returns
    -------
    n : int
        The number of samples necessary to meet the constraints.

    References
    ----------
    .. [2] Hahn, Gerald J., and Meeker, William Q. "Statistical Intervals: A
       Guide for Practitioners." Germany, Wiley, 1991.
    """
    order_stat_var_check(p=P, k=k, c=c, nmax=nmax)

    # use bisection to get minimum n (secant method is unstable due to flat portions of curve)
    nmin = np.ceil(max(k/P - 1, k/(1-P) - 1))
    ntemp = nmin
    n = [nmin, nmax]
    maxsteps = 100  # nmax hard limit of 2^100

    (iPl, iP, iPu) = get_iP(n[0], P)
    if bound == StatBound.TWOSIDED:
        l = iPl - k + 1  # we won't be using assymmetrical order stats
        u = iPu + k - 1
        if l <= 0 or u >= n[1] + 1 or EPYP(n[0], l, u, P) < c:
            raise ValueError(f'n ouside bounds of {nmin=}:{nmax=} for {P=} with {k=} ' +
                             f'at {c=}. Increase nmax, raise k, or loosen constraints.')
    elif bound == StatBound.ONESIDED_UPPER:
        l = 0
        u = iPu + k - 1
        if u >= n[1] + 1 or EPYP(n[0], l, u, P) < c:
            raise ValueError(f'n ouside bounds of {nmin=}:{nmax=} for {P=} with {k=} ' +
                             f'at {c=}. Increase nmax, raise k, or loosen constraints.')
    elif bound == StatBound.ONESIDED_LOWER:
        l = iPl - k + 1
        u = n[0] + 1
        if l <= 0 or EPYP(n[0], l, u, P) < c:
            raise ValueError(f'n ouside bounds of {nmin=}:{nmax=} for {P=} with {k=} ' +
                             f'at {c=}. Increase nmax, raise k, or loosen constraints.')
    else:
        raise ValueError(f'{bound=} must be {StatBound.ONESIDED_UPPER}, ' +
                         f'{StatBound.ONESIDED_LOWER}, or {StatBound.TWOSIDED}')

    for i in range(maxsteps):
        step = (n[1]-n[0])/2
        if step < 1:
            return int(n[0])
        else:
            ntemp = n[0] + np.ceil(step)
            (iPl, iP, iPu) = get_iP(ntemp, P)
            if bound == StatBound.TWOSIDED:
                l = iPl - k  # we won't be using assymmetrical order stats
                u = iPu + k
            elif bound == StatBound.ONESIDED_UPPER:
                l = 0
                u = iPu + k
            elif bound == StatBound.ONESIDED_LOWER:
                l = iPl - k
                u = ntemp + 1
            if EPYP(ntemp, l, u, P) > c:
                n[0] = ntemp
            else:
                n[1] = ntemp
        # print(ntemp, ':', EPYP(ntemp, l, u, P), l, iP, u, n, step)
    raise ValueError(f'With {n=}, could not converge in {maxsteps=} steps. ' +
                     f'Is {nmax=} > 2^{maxsteps}?')



def order_stat_P_k(n     : int,
                   P     : float,
                   c     : float,
                   bound : StatBound = StatBound.TWOSIDED,
                   ) -> int:
    """
    For an Order Statistic Percentile, find the maximum p from n, k, and c.

    Parameters
    ----------
    n : int
        The number of samples.
    P : float (0 < P < 1)
        The target percentile.
    c : float (0 < c < 1)
        The confidence of the interval bound.
    ptol : float, default: 1e-9
        The absolute tolerance on determining p.
    bound : monaco.mc_enums.StatBound, default: '2-sided'
        The statistical bound, '1-sided upper', '1-sided lower', or '2-sided'.

    Returns
    -------
    k : int
        The k'th order statistic meeting the input constraints.
    """
    order_stat_var_check(n=n, p=P, c=c)

    (iPl, iP, iPu) = get_iP(n, P)
    if bound == StatBound.TWOSIDED:
        k = [1, min(iPl, n + 1 - iPu)]
        l = iPl - k[1] + 1  # we won't be using assymmetrical order stats
        u = iPu + k[1] - 1
        if l <= 0 or u >= n+1 or EPYP(n, l, u, P) < c:
            raise ValueError(f'{n=} is too small to meet {P=} at {c=} for {bound} percentile ' +
                              'confidence interval at any order statistic')

    elif bound == StatBound.ONESIDED_UPPER:
        k = [1, n + 1 - iPu]
        l = 0
        u = iPu + k[1] - 1
        if u >= n + 1 or EPYP(n, l, u, P) < c:
            raise ValueError(f'{n=} is too small to meet {P=} at {c=} for {bound} percentile ' +
                              'confidence interval at any order statistic')

    elif bound == StatBound.ONESIDED_LOWER:
        k = [1, iPl]
        l = iPl - k[1] + 1
        u = n + 1
        if EPYP(n, l, u, P) < c:
            raise ValueError(f'{n=} is too small to meet {P=} at {c=} for {bound} percentile ' +
                              'confidence interval at any order statistic')
    else:
        raise ValueError(f'{bound=} must be {StatBound.ONESIDED_UPPER}, ' +
                         f'{StatBound.ONESIDED_LOWER}, or {StatBound.TWOSIDED}')

    # use bisection to get n (secant method is unstable due to flat portions of curve)
    maxsteps = 100  # nmax hard limit of 2^100
    for _ in range(maxsteps):
        step = (k[1]-k[0])/2
        ktemp = k[0] + np.ceil(step)

        if step < 1:
            return int(k[1])

        else:
            if bound == StatBound.TWOSIDED:
                l = iPl - ktemp
                u = iPu + ktemp
            elif bound == StatBound.ONESIDED_UPPER:
                l = 0
                u = iPu + ktemp
            elif bound == StatBound.ONESIDED_LOWER:
                l = iPl - ktemp
                u = n + 1

            if EPYP(n, l, u, P) > c:
                k[1] = ktemp
            else:
                k[0] = ktemp

    raise ValueError(f'With {n=}, could not converge in {maxsteps} steps. Is n > 2^{maxsteps}?')



def order_stat_P_c(n     : int,
                   k     : int,
                   P     : float,
                   bound : StatBound = StatBound.TWOSIDED,
                   ) -> float:
    """
    For an Order Statistic percentile, find the maximum c from n, k, and P.

    Parameters
    ----------
    n : int
        The number of samples.
    k : int
        The k'th order statistic.
    P : float (0 < P < 1)
        The target percentile.
    bound : monaco.mc_enums.StatBound, default: '2-sided'
        The statistical bound, '1-sided upper', '1-sided lower', or '2-sided'.

    Returns
    -------
    c : float (0 < c < 1)
        The confidence of the interval bound.
    """
    order_stat_var_check(n=n, p=P, k=k)

    (iPl, iP, iPu) = get_iP(n, P)
    if bound == StatBound.TWOSIDED:
        l = iPl - k  # we won't be using assymmetrical order stats
        u = iPu + k
    elif bound == StatBound.ONESIDED_UPPER:
        l = 0
        u = iPu + k
    elif bound == StatBound.ONESIDED_LOWER:
        l = iPl - k
        u = n + 1
    else:
        raise ValueError(f'{bound=} must be {StatBound.ONESIDED_UPPER}, ' +
                         f'{StatBound.ONESIDED_LOWER}, or {StatBound.TWOSIDED}')

    if l < 0 or u > n+1:
        raise ValueError(f'{l=} or {u=} are outside the valid bounds of (0, {n+1}) ' +
                         f'(check: {iP=}, {k=})')

    c = EPYP(n, l, u, P)
    return c



def EPYP(n : int,
         l : int,
         u : int,
         p : float,
         ) -> float:
    """
    Estimated Probabiliity for the Y'th Percentile, see Chp. 5.2 of Reference.

    Parameters
    ----------
    n : int
        TODO Description
    l : int
        TODO Description
    u : int
        TODO Description
    p : float (0 < p < 1)
        TODO Description

    Returns
    -------
    c : float (0 < c < 1)
        TODO Description
    """
    order_stat_var_check(n=n, l=l, u=u, p=p)
    c = scipy.stats.binom.cdf(u-1, n, p) - scipy.stats.binom.cdf(l-1, n, p)
    return c



def EPTI(n : int,
         l : int,
         u : int,
         p : float,
         ) -> float:
    """
    Estimated Probabiliity for a Tolerance Interval, see Chp. 5.3 of Reference

    Parameters
    ----------
    n : int
        TODO Description
    l : int
        TODO Description
    u : int
        TODO Description
    p : float (0 < p < 1)
        TODO Description

    Returns
    -------
    c : float (0 < c < 1)
        TODO Description
    """
    order_stat_var_check(n=n, l=l, u=u, p=p)
    c = scipy.stats.binom.cdf(u-l-1, n, p)
    return c



def get_iP(n : int,
           P : float,
           ) -> tuple[int, int, int]:
    """
    Get the index of Percentile (1-based indexing)

    Parameters
    ----------
    n : int
        Number of samples
    P : float (0 < P < 1)
        Target percentile

    Returns
    -------
    (iPl, iP, iPu) : (int, int, int)
        Lower, closest, and upper index of the percentile.
    """
    iP = P*(n + 1)
    iPl = int(np.floor(iP))
    iPu = int(np.ceil(iP))
    iP = int(np.round(iP))
    return (iPl, iP, iPu)



def order_stat_var_check(n    : int | None   = None,
                         l    : int | None   = None,
                         u    : int | None   = None,
                         p    : float | None = None,
                         k    : int | None   = None,
                         c    : float | None = None,
                         nmax : int | None   = None
                         ) -> None:
    """
    Check the validity of the inputs to the order statistic functions.
    """
    if n is not None and n < 1:
        raise ValueError(f'{n=} must be >= 1')
    if l is not None and l < 0:
        raise ValueError(f'{l=} must be >= 0')
    if u is not None and n is not None and u > n+1:
        raise ValueError(f'{u=} must be >= {n+1}')
    if u is not None and l is not None and u < l:
        raise ValueError(f'{u=} must be >= {l=}')
    if p is not None and (p <= 0 or p >= 1):
        raise ValueError(f'{p=} must be in the range 0 < p < 1')
    if k is not None and k < 1:
        raise ValueError(f'{k=} must be >= 1')
    if c is not None and (c <= 0 or c >= 1):
        raise ValueError(f'{c=} must be in the range 0 < c < 1')
    if nmax is not None and nmax < 1:
        raise ValueError(f'{nmax=} must be >= 1')
